

import os
from typing import Any

from nbconvert.exporters import HTMLExporter
from nbconvert.preprocessors import CSSHTMLHeaderPreprocessor
from traitlets import List, default, Bool
from traitlets.config import Config

from grader_service.convert.converters.baseapp import ConverterApp
from grader_service.convert.preprocessors import GetGrades
from grader_service.convert.converters.base import BaseConverter


class GenerateFeedback(BaseConverter):
    preprocessors = List([GetGrades, CSSHTMLHeaderPreprocessor]).tag(config=True)

    @default("classes")
    def _classes_default(self):
        classes = super(GenerateFeedback, self)._classes_default()
        classes.append(HTMLExporter)
        return classes

    @default("export_class")
    def _exporter_class_default(self):
        return HTMLExporter

    @default("permissions")
    def _permissions_default(self):
        return 664

    def __init__(
        self, input_dir: str, output_dir: str, file_pattern: str, **kwargs: Any
    ):
        super(GenerateFeedback, self).__init__(
            input_dir, output_dir, file_pattern, **kwargs
        )
        c = Config()
        # Note: nbconvert 6.0 completely changed how templates work: they can now be installed separately
        #  and can be given by name (classic is default)
        if "template" not in self.config.HTMLExporter:
            template_path = os.path.abspath(
                os.path.join(
                    os.path.dirname(__file__),
                    "..",
                    "templates",
                )
            )
            # '/Users/matthiasmatt/opt/miniconda3/envs/grader/share/jupyter/nbconvert/templates/classic/index.html.j2'
            c.TemplateExporter.extra_template_basedirs = template_path
            c.HTMLExporter.template_name = "feedback"
        self.update_config(c)
        self.force = True  # always overwrite generated assignments


class GenerateFeedbackApp(ConverterApp):
    version = ConverterApp.__version__

    copy_files = Bool(False, allow_none=False).tag(config=False)

    def start(self):
        GenerateFeedback(
            input_dir=self.input_directory,
            output_dir=self.output_directory,
            file_pattern=self.file_pattern,
            copy_files=self.copy_files,
            config=self.config
        ).start()





from textwrap import dedent
from typing import Any

from traitlets import Bool, List, default
from traitlets.config.loader import Config

from grader_service.convert.converters.baseapp import ConverterApp
from grader_service.convert.preprocessors import (
    ClearMarkScheme,
    ClearOutput,
    Execute,
    IncludeHeaderFooter,
    LockCells,
)
from grader_service.convert.converters.base import BaseConverter


class GenerateSolution(BaseConverter):

    create_assignment = Bool(
        True,
        help=dedent(
            """
            Whether to create the assignment at runtime if it does not
            already exist.
            """
        ),
    ).tag(config=True)

    @default("permissions")
    def _permissions_default(self) -> int:
        return 664

    def _load_config(self, cfg: Config, **kwargs: Any) -> None:
        super(GenerateSolution, self)._load_config(cfg, **kwargs)

    preprocessors = List(
        [IncludeHeaderFooter, LockCells, ClearOutput, ClearMarkScheme, Execute]
    )
    # NB: ClearHiddenTests must come after ComputeChecksums and SaveCells.
    # ComputerChecksums must come again after ClearHiddenTests.

    def __init__(
        self, input_dir: str, output_dir: str, file_pattern: str, **kwargs: Any
    ) -> None:
        super(GenerateSolution, self).__init__(
            input_dir, output_dir, file_pattern, **kwargs
        )
        self.force = True  # always overwrite generated assignments

    def start(self) -> None:
        super(GenerateSolution, self).start()


class GenerateSolutionApp(ConverterApp):
    version = ConverterApp.__version__

    def start(self):
        GenerateSolution(
            input_dir=self.input_directory,
            output_dir=self.output_directory,
            file_pattern=self.file_pattern,
            copy_files=self.copy_files,
            config=self.config
        ).start()





from typing import Tuple

from nbconvert.exporters.exporter import ResourcesDict
from nbformat.notebooknode import NotebookNode
from traitlets import Bool

from grader_service.convert import utils
from grader_service.convert.preprocessors.base import NbGraderPreprocessor


class LockCells(NbGraderPreprocessor):
    """A preprocessor for making cells undeletable."""

    lock_solution_cells = Bool(
        True, help="Whether solution cells are locked (non-deletable and non-editable)"
    ).tag(config=True)

    lock_grade_cells = Bool(
        True, help="Whether grade cells are locked (non-deletable)"
    ).tag(config=True)

    lock_readonly_cells = Bool(
        True, help="Whether readonly cells are locked (non-deletable and non-editable)"
    ).tag(config=True)

    lock_all_cells = Bool(
        False,
        help="Whether all assignment cells are locked (non-deletable and non-editable)",
    ).tag(config=True)

    def preprocess_cell(
        self, cell: NotebookNode, resources: ResourcesDict, cell_index: int
    ) -> Tuple[NotebookNode, ResourcesDict]:
        if (
            (self.lock_solution_cells or self.lock_grade_cells)
            and utils.is_solution(cell)
            and utils.is_grade(cell)
        ):
            cell.metadata["deletable"] = False
        elif self.lock_solution_cells and utils.is_solution(cell):
            cell.metadata["deletable"] = False
        elif self.lock_grade_cells and utils.is_grade(cell):
            cell.metadata["deletable"] = False
            cell.metadata["editable"] = False
        elif self.lock_readonly_cells and utils.is_locked(cell):
            cell.metadata["deletable"] = False
            cell.metadata["editable"] = False
        elif self.lock_all_cells:
            cell.metadata["deletable"] = False
            cell.metadata["editable"] = False
        return cell, resources



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from sqlalchemy import Column, String, LargeBinary, Unicode
from sqlalchemy.orm import relationship, Session

from grader_service.api.models import user
from grader_service.auth.crypto import encrypt, decrypt, InvalidToken, EncryptionUnavailable, CryptKeeper
from grader_service.orm.base import Base, Serializable
from grader_service.orm.group import group_assignment_table
from grader_service.utils import new_token


class User(Base, Serializable):
    __tablename__ = "user"
    name = Column(String(255), primary_key=True)

    roles = relationship("Role", back_populates="user")
    submissions = relationship("Submission", back_populates="user")
    groups = relationship("Group", secondary=group_assignment_table,
                          back_populates="users")
    api_tokens = relationship("APIToken", back_populates="user")
    oauth_codes = relationship("OAuthCode", back_populates="user")

    encrypted_auth_state = Column(LargeBinary)
    cookie_id = Column(Unicode(255), default=new_token, nullable=False, unique=True)
    _auth_refreshed = None

    async def save_auth_state(self, auth_state):
        if auth_state is None:
            self.encrypted_auth_state = None
        else:
            self.encrypted_auth_state = await encrypt(auth_state)
        session: Session = Session.object_session(self)
        session.expire_on_commit = False
        session.commit()

    async def get_auth_state(self):
        """Retrieve and decrypt auth_state for the user"""
        encrypted = self.encrypted_auth_state
        if encrypted is None:
            return None
        try:
            auth_state = await decrypt(encrypted)
        except (ValueError, InvalidToken, EncryptionUnavailable) as e:
            self.log.warning(
                "Failed to retrieve encrypted auth_state for %s because %s",
                self.name,
                e,
            )
            return
        # loading auth_state
        if auth_state:
            # Crypt has multiple keys, store again with new key for rotation.
            if len(CryptKeeper.instance().keys) > 1:
                await self.save_auth_state(auth_state)
        return auth_state

    def serialize(self):
        return {"name": self.name}


    @property
    def model(self) -> user.User:
        return user.User(name=self.name)



"""Authorization handlers"""
# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
import json
from abc import ABC
from datetime import datetime
from typing import Optional, Awaitable
from unittest import mock
from urllib.parse import parse_qsl, quote, urlencode, urlparse, urlunparse

from oauthlib import oauth2
from tornado import web

from grader_service.orm.api_token import APIToken
from grader_service.utils import get_browser_protocol, token_authenticated, url_path_join, utcnow
from grader_service.handlers.base_handler import BaseHandler


class SelfAPIHandler(BaseHandler):
    """Return the authenticated user's model

    Based on the authentication info. Acts as a 'whoami' for auth tokens.
    """

    async def get(self):
        user = self.current_user
        if user is None:
            raise web.HTTPError(403)
        model = user.model.to_dict()

        # add session_id associated with token
        # added in 2.0
        # token_id added in 5.0
        token = self.get_token()
        if token:
            model["token_id"] = token.api_id
            model["session_id"] = token.session_id
        else:
            model["token_id"] = None
            model["session_id"] = None
        model["roles"] = [f"{r.lecture.code}:{r.role.name}" for r in user.roles]
        self.write(json.dumps(model))


class OAuthHandler:
    def extract_oauth_params(self):
        """extract oauthlib params from a request

        Returns:

        (uri, http_method, body, headers)
        """
        return (
            self.request.uri,
            self.request.method,
            self.request.body,
            self.request.headers,
        )

    def make_absolute_redirect_uri(self, uri):
        """Make absolute redirect URIs

        internal redirect uris, e.g. `/user/foo/oauth_handler`
        are allowed in jupyterhub, but oauthlib prohibits them.
        Add `$HOST` header to redirect_uri to make them acceptable.

        Currently unused in favor of monkeypatching
        oauthlib.is_absolute_uri to skip the check
        """
        redirect_uri = self.get_argument('redirect_uri')
        if not redirect_uri or not redirect_uri.startswith('/'):
            return uri
        # make absolute local redirects full URLs
        # to satisfy oauthlib's absolute URI requirement
        redirect_uri = (
                get_browser_protocol(self.request)
                + "://"
                + self.request.host
                + redirect_uri
        )
        parsed_url = urlparse(uri)
        query_list = parse_qsl(parsed_url.query, keep_blank_values=True)
        for idx, item in enumerate(query_list):
            if item[0] == 'redirect_uri':
                query_list[idx] = ('redirect_uri', redirect_uri)
                break

        return urlunparse(urlparse(uri)._replace(query=urlencode(query_list)))

    def add_credentials(self, credentials=None):
        """Add oauth credentials

        Adds user, session_id, client to oauth credentials
        """
        if credentials is None:
            credentials = {}
        else:
            credentials = credentials.copy()

        session_id = self.get_session_cookie()
        if session_id is None:
            session_id = self.set_session_cookie()

        user = self.current_user

        # Extra credentials we need in the validator
        credentials.update({'user': user, 'handler': self, 'session_id': session_id})
        return credentials

    def send_oauth_response(self, headers, body, status):
        """Send oauth response from provider return values

        Provider methods return headers, body, and status
        to be set on the response.

        This method applies these values to the Handler
        and sends the response.
        """
        self.set_status(status)
        for key, value in headers.items():
            self.set_header(key, value)
        if body:
            self.write(body)


class OAuthAuthorizeHandler(OAuthHandler, BaseHandler):
    """Implement OAuth authorization endpoint(s)"""

    def prepare(self) -> Optional[Awaitable[None]]:
        return super().prepare()

    def _complete_login(self, uri, headers, scopes, credentials):
        try:
            headers, body, status = self.application.oauth_provider.create_authorization_response(
                uri, 'POST', '', headers, scopes, credentials
            )

        except oauth2.FatalClientError as e:
            # TODO: human error page
            raise
        self.send_oauth_response(headers, body, status)

    def needs_oauth_confirm(self, user, oauth_client, requested_scopes):
        """Return whether the given oauth client needs to prompt for access for the given user

        Checks list for oauth clients that don't need confirmation

        Sources:

        - the user's own servers
        - Clients which already have authorization for the same roles
        - Explicit oauth_no_confirm_list configuration (e.g. admin-operated services)

        .. versionadded: 1.1
        """
        # get the oauth client ids for the user's own server(s)
        own_oauth_client_ids = {
            spawner.oauth_client_id for spawner in user.spawners.values()
        }
        if (
                # it's the user's own server
                oauth_client.identifier in own_oauth_client_ids
                # or it's in the global no-confirm list
                or oauth_client.identifier
                in self.settings.get('oauth_no_confirm_list', set())
        ):
            return False

        # Check existing authorization
        existing_tokens = self.session.query(APIToken).filter_by(
            username=user.name,
            client_id=oauth_client.identifier,
        )
        authorized_scopes = set()
        for token in existing_tokens:
            authorized_scopes.update(token.scopes)

        if authorized_scopes:
            if set(requested_scopes).issubset(authorized_scopes):
                self.log.debug(
                    f"User {user.name} has already authorized {oauth_client.identifier} for scopes {requested_scopes}"
                )
                return False
            else:
                self.log.debug(
                    f"User {user.name} has authorized {oauth_client.identifier}"
                    f" for scopes {authorized_scopes}, confirming additional scopes {requested_scopes}"
                )
        # default: require confirmation
        return True

    def get_login_url(self):
        """
        Support automatically logging in when JupyterHub is used as auth provider
        """
        if self.authenticator.auto_login_oauth2_authorize:
            return self.authenticator.login_url(self.application.base_url)
        return super().get_login_url()

    async def get(self):
        """GET /oauth/authorization

        Render oauth confirmation page:
        "Server at ... would like permission to ...".

        Users accessing their own server or a blessed service
        will skip confirmation.
        """
        uri, http_method, body, headers = self.extract_oauth_params()
        try:
            with mock.patch.object(
                self.oauth_provider.request_validator,
                "_current_user",
                self.current_user,
                create=True,
            ):
                (
                    requested_scopes,
                    credentials,
                ) = self.oauth_provider.validate_authorization_request(
                    uri, http_method, body, headers
                )
            credentials = self.add_credentials(credentials)
            client = self.oauth_provider.fetch_by_client_id(credentials['client_id'])

            # Render oauth 'Authorize application...' page
            auth_state = await self.current_user.get_auth_state()
            self.write(
                await self.render_template(
                    "auth/oauth.html.j2",
                    auth_state=auth_state,
                    oauth_client=client,
                )
            )

        # Errors that should be shown to the user on the provider website
        except oauth2.FatalClientError as e:
            raise web.HTTPError(e.status_code, e.description)

        # Errors embedded in the redirect URI back to the client
        except oauth2.OAuth2Error as e:
            self.log.error("OAuth error: %s", e.description)
            self.redirect(e.in_uri(e.redirect_uri))

    def post(self):
        uri, http_method, body, headers = self.extract_oauth_params()
        # The scopes the user actually authorized, i.e. checkboxes
        # that were selected.
        scopes = self.get_arguments('scopes')
        if scopes == []:
            # avoid triggering default scopes (provider selects default scopes when scopes is falsy)
            # when an explicit empty list is authorized
            scopes = ["identify"]
        # credentials we need in the validator
        credentials = self.add_credentials()

        try:
            headers, body, status = self.application.oauth_provider.create_authorization_response(
                uri, http_method, body, headers, scopes, credentials
            )
        except oauth2.FatalClientError as e:
            raise web.HTTPError(e.status_code, e.description)
        else:
            self.send_oauth_response(headers, body, status)


class OAuthTokenHandler(OAuthHandler, BaseHandler):
    def post(self):
        uri, http_method, body, headers = self.extract_oauth_params()
        credentials = {}

        try:
            headers, body, status = self.application.oauth_provider.create_token_response(
                uri, http_method, body, headers, credentials
            )
        except oauth2.FatalClientError as e:
            raise web.HTTPError(e.status_code, e.description)
        else:
            self.send_oauth_response(headers, body, status)


def get_oauth_default_handlers(base_path: str):
    return [
        (url_path_join(base_path, "/api/user"), SelfAPIHandler),
        (url_path_join(base_path, "/api/oauth2/authorize"), OAuthAuthorizeHandler),
        (url_path_join(base_path, "/api/oauth2/token"), OAuthTokenHandler),
    ]



from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from grader_service.api.models.base_model import Model
from grader_service.api import util


class SubmissionPeriod(Model):
    """NOTE: This class is auto generated by OpenAPI Generator (https://openapi-generator.tech).

    Do not edit the class manually.
    """

    def __init__(self, period=None, scaling=None):  # noqa: E501
        """SubmissionPeriod - a model defined in OpenAPI

        :param period: The period of this SubmissionPeriod.  # noqa: E501
        :type period: str
        :param scaling: The scaling of this SubmissionPeriod.  # noqa: E501
        :type scaling: float
        """
        self.openapi_types = {
            'period': str,
            'scaling': float
        }

        self.attribute_map = {
            'period': 'period',
            'scaling': 'scaling'
        }

        self._period = period
        self._scaling = scaling

    @classmethod
    def from_dict(cls, dikt) -> 'SubmissionPeriod':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The SubmissionPeriod of this SubmissionPeriod.  # noqa: E501
        :rtype: SubmissionPeriod
        """
        return util.deserialize_model(dikt, cls)

    @property
    def period(self) -> str:
        """Gets the period of this SubmissionPeriod.


        :return: The period of this SubmissionPeriod.
        :rtype: str
        """
        return self._period

    @period.setter
    def period(self, period: str):
        """Sets the period of this SubmissionPeriod.


        :param period: The period of this SubmissionPeriod.
        :type period: str
        """

        self._period = period

    @property
    def scaling(self) -> float:
        """Gets the scaling of this SubmissionPeriod.


        :return: The scaling of this SubmissionPeriod.
        :rtype: float
        """
        return self._scaling

    @scaling.setter
    def scaling(self, scaling: float):
        """Sets the scaling of this SubmissionPeriod.


        :param scaling: The scaling of this SubmissionPeriod.
        :type scaling: float
        """

        self._scaling = scaling



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
# grader_s/grader_s/handlers
import json
import logging
import shutil
import time

from grader_service.orm.base import DeleteState
import isodate
import jwt
import os.path
import subprocess
from http import HTTPStatus
import tornado
from celery import chain

from grader_service.plugins.lti import LTISyncGrades
from grader_service.autograding.local_feedback import GenerateFeedbackExecutor
from grader_service.handlers.handler_utils import parse_ids
from grader_service.api.models.submission import Submission as SubmissionModel
from grader_service.api.models.assignment_settings import AssignmentSettings as AssignmentSettingsModel
from grader_service.orm.assignment import AutoGradingBehaviour
from grader_service.orm.assignment import Assignment
from grader_service.orm.lecture import Lecture
from grader_service.orm.submission import Submission
from grader_service.orm.submission_logs import SubmissionLogs
from grader_service.orm.submission_properties import SubmissionProperties
from grader_service.orm.takepart import Role, Scope
from grader_service.registry import VersionSpecifier, register_handler
from sqlalchemy.sql.expression import func
from tornado.web import HTTPError
from grader_service.convert.gradebook.models import GradeBookModel
from grader_service.autograding.celery.tasks import autograde_task, generate_feedback_task, lti_sync_task

from grader_service.handlers.base_handler import GraderBaseHandler, authorize, \
    RequestHandlerConfig
import datetime


def remove_points_from_submission(submissions):
    for s in submissions:
        if s.feedback_status not in ('generated', 'feedback_outdated'):
            s.score = None
    return submissions


@register_handler(
    path=r'\/api\/lectures\/(?P<lecture_id>\d*)\/assignments' +
         r'\/(?P<assignment_id>\d*)\/submissions\/?',
    version_specifier=VersionSpecifier.ALL,
)
class SubmissionHandler(GraderBaseHandler):
    """Tornado Handler class for http requests to
    /lectures/{lecture_id}/assignments/{assignment_id}/submissions.
    """

    def on_finish(self):
        # we do not close the session we just commit because we might run
        # LocalAutogradeExecutor or GenerateFeedbackExecutor in POST which
        # still need it
        pass

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int, assignment_id: int):
        """Return the submissions of an assignment.

        Two query parameter: latest, instructor-version.

        latest: only get the latest submissions of users.
        instructor-version: if true, get the submissions of all users in
        lecture if false, get own submissions.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :raises HTTPError: throws err if user is not authorized or
        the assignment was not found
        """
        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)
        self.validate_parameters("filter", "instructor-version", "format")
        submission_filter = self.get_argument("filter", "none")
        if submission_filter not in ["none", "latest", "best"]:
            raise HTTPError(HTTPStatus.BAD_REQUEST, reason="Filter parameter \
            has to be either 'none', 'latest' or 'best'")
        instr_version = self.get_argument("instructor-version", None) == "true"
        response_format = self.get_argument("format", "json")
        if response_format not in ["json", "csv"]:
            raise HTTPError(HTTPStatus.BAD_REQUEST, reason="Response format \
            can either be 'json' or 'csv'")

        role: Role = self.get_role(lecture_id)
        if instr_version and role.role < Scope.tutor:
            raise HTTPError(HTTPStatus.FORBIDDEN, reason="Forbidden")
        assignment = self.get_assignment(lecture_id, assignment_id)

        if instr_version:
            if submission_filter == 'latest':

                # build the subquery
                subquery = (
                    self.session.query(Submission.username,
                                       func.max(Submission.date).label(
                                           "max_date"))
                    .filter(Submission.assignid == assignment_id)
                    .filter(Submission.deleted == DeleteState.active)
                    .group_by(Submission.username)
                    .subquery())

                # build the main query
                submissions = (
                    self.session.query(Submission)
                    .join(subquery,
                          (Submission.username == subquery.c.username) & (
                                  Submission.date == subquery.c.max_date) & (
                                  Submission.assignid == assignment_id) & (
                                  Submission.deleted == DeleteState.active
                                  ))
                    .order_by(Submission.id)
                    .all())

            elif submission_filter == 'best':

                # build the subquery
                subquery = (
                    self.session.query(Submission.username,
                                       func.max(Submission.score).label(
                                           "max_score"))
                    .filter(Submission.assignid == assignment_id)
                    .filter(Submission.deleted == DeleteState.active)
                    .group_by(Submission.username)
                    .subquery())

                # build the main query
                submissions = (
                    self.session.query(Submission)
                    .join(subquery,
                          (Submission.username == subquery.c.username) & (
                                  Submission.score == subquery.c.max_score) & (
                                  Submission.assignid == assignment_id) & (
                                  Submission.deleted == DeleteState.active
                                  ))
                    .order_by(Submission.id)
                    .all())

            else:
                submissions = [
                s for s in assignment.submissions if (s.deleted
                                                        == DeleteState.active)
                ]
        else:
            if submission_filter == 'latest':
                # build the subquery
                subquery = (self.session.query(Submission.username,
                                               func.max(Submission.date).label(
                                                   "max_date"))
                            .filter(Submission.assignid == assignment_id)
                            .filter(Submission.deleted == DeleteState.active)
                            .group_by(Submission.username)
                            .subquery())

                # build the main query
                submissions = (
                    self.session.query(Submission)
                    .join(subquery,
                          (Submission.username == subquery.c.username) & (
                                  Submission.date == subquery.c.max_date) & (
                                  Submission.assignid == assignment_id) & (
                                  Submission.deleted == DeleteState.active
                                  ))
                    .filter(
                        Submission.assignid == assignment_id,
                        Submission.username == role.username, )
                    .order_by(Submission.id)
                    .all())

            elif submission_filter == 'best':

                # build the subquery
                subquery = (self.session.query(Submission.username, func.max(
                    Submission.score).label("max_score"))
                            .filter(Submission.assignid == assignment_id)
                            .filter(Submission.deleted == DeleteState.active)
                            .group_by(Submission.username)
                            .subquery())

                # build the main query
                submissions = (
                    self.session.query(Submission)
                    .join(subquery,
                          (Submission.username == subquery.c.username) & (
                                  Submission.score == subquery.c.max_score) & (
                                  Submission.assignid == assignment_id) & (
                                  Submission.deleted == DeleteState.active
                                  ))
                    .filter(
                        Submission.assignid == assignment_id,
                        Submission.username == role.username, )
                    .order_by(Submission.id)
                    .all())
            else:
                submissions = (
                    self.session.query(
                        Submission
                    )
                    .filter(
                        Submission.assignid == assignment_id,
                        Submission.username == role.username,
                        Submission.deleted == DeleteState.active
                    )
                    .order_by(Submission.id)
                    .all()
                )
        if response_format == "csv":
            # csv format does not include logs
            self.set_header("Content-Type", "text/csv")
            for i, s in enumerate(submissions):
                d = s.model.to_dict()
                if i == 0:
                    self.write(
                        ",".join((k for k in d.keys() if k != "logs")) + "\n")
                self.write(",".join(
                    (str(v) for k, v in d.items() if k != "logs")) + "\n")
        else:
            if not instr_version:
                submissions = remove_points_from_submission(submissions)

            self.write_json(submissions)
        self.session.close()  # manually close here because on_finish overwrite

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def post(self, lecture_id: int, assignment_id: int):
        """Create submission based on commit hash.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :raises HTTPError: throws err if user is not authorized or
        the assignment was not found
        """
        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)
        self.validate_parameters()
        body = tornado.escape.json_decode(self.request.body)
        try:
            commit_hash = body["commit_hash"]
        except KeyError:
            raise HTTPError(400, reason="Commit hash not found in body")

        role = self.get_role(lecture_id)
        assignment = self.get_assignment(lecture_id, assignment_id)
        if assignment.status == "complete":
            raise HTTPError(HTTPStatus.BAD_REQUEST,
                            reason="Cannot submit completed assignment!")
        if role.role == Scope.student and assignment.status != "released":
            raise HTTPError(HTTPStatus.NOT_FOUND)
        # set utc time
        submission_ts = datetime.datetime.now(datetime.timezone.utc)
        # use implicit utc time to compare with database objects
        submission_ts = submission_ts.replace(tzinfo=None)
        

        score_scaling = 1.0
        if assignment.duedate is not None:
            score_scaling = self.calculate_late_submission_scaling(assignment, submission_ts, role)

        if assignment.max_submissions:
            submissions = assignment.submissions
            usersubmissions = [s for s in submissions if
                               s.username == role.username]
            if len(usersubmissions) >= assignment.max_submissions and role.role < Scope.tutor:
                raise HTTPError(HTTPStatus.CONFLICT, reason="Maximum number \
                of submissions reached!")

        submission = Submission()
        submission.assignid = assignment.id
        submission.date = submission_ts
        submission.username = self.user.name
        submission.score_scaling = score_scaling

        git_repo_path = self.construct_git_dir(
            repo_type=assignment.type, lecture=assignment.lecture,
            assignment=assignment)
        if git_repo_path is None or not os.path.exists(git_repo_path):
            raise HTTPError(HTTPStatus.NOT_FOUND,
                            reason="Git repository not found")

        try:
            # Commit hash "0"*40 is used to differentiate between submissions created by instructors for students and normal submissions by any user.
            # In this case submissions for the student might not exist, so we cannot reference a non-existing commit_hash.
            # When submission is set to editted, autograder uses edit repository, so we don't need the commit_hash of the submission.
            if commit_hash != "0" * 40:
                subprocess.run(
                    ["git", "branch", "main", "--contains", commit_hash],
                    cwd=git_repo_path, capture_output=True)
        except subprocess.CalledProcessError:
            raise HTTPError(HTTPStatus.NOT_FOUND, reason="Commit not found")

        submission.commit_hash = commit_hash
        submission.auto_status = "not_graded"
        submission.manual_status = "not_graded"
        submission.feedback_status = "not_generated"

        automatic_grading = assignment.automatic_grading

        self.session.add(submission)
        self.session.commit()
        self.set_status(HTTPStatus.CREATED)
        self.write_json(submission)

        # If the assignment has automatic grading or fully
        # automatic grading perform necessary operations
        if automatic_grading in [AutoGradingBehaviour.auto,
                                 AutoGradingBehaviour.full_auto]:
            submission.auto_status = "pending"
            self.session.commit()
            self.set_status(HTTPStatus.ACCEPTED)

            if automatic_grading == AutoGradingBehaviour.full_auto:
                submission.feedback_status = "generating"
                self.session.commit()

                # use immutable signature: https://docs.celeryq.dev/en/stable/reference/celery.app.task.html#celery.app.task.Task.si
                grading_chain = chain(
                    autograde_task.si(lecture_id, assignment_id, submission.id),
                    generate_feedback_task.si(lecture_id, assignment_id, submission.id),
                    lti_sync_task.si(lecture_id, assignment_id, submission.id, feedback_sync=True)
                )
            else:
                grading_chain = chain(autograde_task.si(lecture_id, assignment_id, submission.id))
            grading_chain()

        if automatic_grading == AutoGradingBehaviour.unassisted:
            self.session.close()


    @staticmethod
    def calculate_late_submission_scaling(assignment, submission_ts, role: Role) -> float:
        assignment_settings = AssignmentSettingsModel.from_dict(json.loads(assignment.settings))
        # make duedate aware
        if assignment_settings.late_submission and len(assignment_settings.late_submission) > 0:
            scaling = 0.0
            if submission_ts <= assignment.duedate:
                scaling = 1.0
            else:
                for period in assignment_settings.late_submission:
                    late_submission_date = assignment.duedate + isodate.parse_duration(period.period)
                    if submission_ts < late_submission_date:
                        scaling = period.scaling
                        break
                if scaling == 0.0 and role.role < Scope.tutor:
                    raise HTTPError(HTTPStatus.CONFLICT,
                                    reason="Submission after last late submission period of assignment!")
        else:
            if submission_ts < assignment.duedate:
                scaling = 1.0
            else:
                if role.role < Scope.tutor:
                    raise HTTPError(HTTPStatus.CONFLICT, reason="Submission after due date of assignment!")
                else:
                    scaling = 0.0
        return scaling


@register_handler(
    path=r'\/api\/lectures\/(?P<lecture_id>\d*)\/assignments\/' +
         r'(?P<assignment_id>\d*)\/submissions\/(?P<submission_id>\d*)\/?',
    version_specifier=VersionSpecifier.ALL,
)
class SubmissionObjectHandler(GraderBaseHandler):
    """Tornado Handler class for http requests to
    /lectures/{lecture_id}/assignments/{assignment_id}/submissions
    /{submission_id}.
    """

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int, assignment_id: int,
                  submission_id: int):
        """Returns a specific submission.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :param submission_id: id of the submission
        :type submission_id: int
        """
        lecture_id, assignment_id, submission_id = parse_ids(
            lecture_id, assignment_id, submission_id
        )
        submission = self.get_submission(lecture_id, assignment_id,
                                         submission_id)
        if self.get_role(lecture_id).role == Scope.student \
                and submission.username != self.user.name:
            raise HTTPError(HTTPStatus.NOT_FOUND)
        self.write_json(submission)

    @authorize([Scope.tutor, Scope.instructor])
    async def put(self, lecture_id: int, assignment_id: int,
                  submission_id: int):
        """Updates a specific submission and returns the updated entity.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :param submission_id: id of the submission
        :type submission_id: int
        """
        lecture_id, assignment_id, submission_id = parse_ids(
            lecture_id, assignment_id, submission_id
        )
        body = tornado.escape.json_decode(self.request.body)
        sub_model = SubmissionModel.from_dict(body)
        sub = self.get_submission(lecture_id, assignment_id, submission_id)
        # sub.date = sub_model.submitted_at
        # sub.assignid = assignment_id

        role = self.get_role(lecture_id)
        if role.role >= Scope.instructor:
            sub.username = sub_model.username
        sub.auto_status = sub_model.auto_status
        sub.manual_status = sub_model.manual_status
        sub.edited = sub_model.edited
        sub.feedback_status = sub_model.feedback_status
        if sub_model.score_scaling is not None and sub.score_scaling != sub_model.score_scaling:
            sub.score_scaling = sub_model.score_scaling
            sub.score = sub_model.score_scaling * sub.grading_score
        self.session.commit()
        self.write_json(sub)


    def delete(self, lecture_id: int, assignment_id: int,
                  submission_id: int):
        """Soft-Deletes a specific submission.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :param submission_id: id of the submission
        :type submission_id: int
        :raises HTTPError: if submission can't be deleted due to it having feedback
        """
        lecture_id, assignment_id, submission_id = parse_ids(lecture_id, assignment_id, submission_id)
        self.validate_parameters()
        submission = self.get_submission(lecture_id, assignment_id, submission_id)

        if submission is not None:
            if submission.feedback_status is not "not_generated":
                raise HTTPError(HTTPStatus.FORBIDDEN, reason="Only submissions without feedback can be deleted.")
            # if assignment has deadline
            if submission.assignment.duedate:
                # if assignment's deadline has passed
                if submission.assignment.duedate < datetime.datetime.now().replace(tzinfo=None):
                    raise HTTPError(HTTPStatus.FORBIDDEN, reason="Submission can't be deleted, due date of assigment "
                                                                 "has passed.")
            else:
                previously_deleted = (
                    self.session.query(Submission)
                    .filter(
                        Submission.id == submission_id,
                        Submission.assignid == assignment_id,
                        Submission.deleted == DeleteState.deleted,
                    )
                    .one_or_none()
                )
                if previously_deleted is not None:
                    self.session.delete(previously_deleted)
                    self.session.commit()

                submission.deleted = DeleteState.deleted
                self.session.commit()
        else:
            raise HTTPError(HTTPStatus.NOT_FOUND,
                            reason="Submission to delete not found.")

@register_handler(
    path=r'\/api\/lectures\/(?P<lecture_id>\d*)\/assignments\/' +
         r'(?P<assignment_id>\d*)\/submissions\/(?P<submission_id>\d*)\/logs\/?',
    version_specifier=VersionSpecifier.ALL,
)
class SubmissionLogsHandler(GraderBaseHandler):
    @authorize([Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int, assignment_id: int,
                  submission_id: int):
        """Returns logs of a submission.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :param submission_id: id of the submission
        :type submission_id: int
        :raises HTTPError: throws err if the submission logs are not found
        """
        lecture_id, assignment_id, submission_id = parse_ids(
            lecture_id, assignment_id, submission_id
        )
        logs = self.session.query(SubmissionLogs).get(submission_id)
        if logs is not None:
            self.write_json(logs.logs)
        else:
            raise HTTPError(HTTPStatus.NOT_FOUND,
                            reason="Properties of submission were not found")


@register_handler(
    path=r'\/api\/lectures\/(?P<lecture_id>\d*)\/assignments\/' +
         r'(?P<assignment_id>\d*)\/submissions\/(?P<submission_id>\d*)\/' +
         r'properties\/?',
    version_specifier=VersionSpecifier.ALL,
)
class SubmissionPropertiesHandler(GraderBaseHandler):
    """Tornado Handler class for http requests to
    /lectures/{lecture_id}/assignments/{assignment_id}/submissions/
    {submission_id}/properties.
    """

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int, assignment_id: int,
                  submission_id: int):
        """Returns the properties of a submission,

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :param submission_id: id of the submission
        :type submission_id: int
        :raises HTTPError: throws err if the submission or
        their properties are not found
        """
        lecture_id, assignment_id, submission_id = parse_ids(
            lecture_id, assignment_id, submission_id
        )
        properties = self.session.get(SubmissionProperties, submission_id)
        if properties is not None and properties.properties is not None:
            # delete source cells from properties if user is student
            if self.get_role(lecture_id).role == Scope.student:
                model = GradeBookModel.from_dict(
                    json.loads(properties.properties))
                for notebook in model.notebooks.values():
                    notebook.source_cells_dict = {}
                self.write(json.dumps(model.to_dict()))
            else:
                self.write(properties.properties)
        else:
            raise HTTPError(HTTPStatus.NOT_FOUND,
                            reason="Properties of submission were not found")

    @authorize([Scope.tutor, Scope.instructor])
    async def put(self, lecture_id: int, assignment_id: int,
                  submission_id: int):
        """Updates the properties of a submission.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :param submission_id: id of the submission
        :type submission_id: int
        :raises HTTPError: throws err if the submission are not found
        """
        lecture_id, assignment_id, submission_id = parse_ids(
            lecture_id, assignment_id, submission_id
        )
        submission = self.get_submission(lecture_id, assignment_id,
                                         submission_id)
        properties_string: str = self.request.body.decode("utf-8")

        try:
            gradebook = GradeBookModel.from_dict(json.loads(properties_string))

            score = gradebook.score
            submission.grading_score = score
            submission.score = submission.score_scaling * score

        except Exception as e:
            self.log.info(e)
            raise HTTPError(HTTPStatus.BAD_REQUEST,
                            reason="Cannot parse properties file!")

        properties = SubmissionProperties(properties=properties_string,
                                          sub_id=submission.id)

        self.session.merge(properties)

        if submission.feedback_status == 'generated':
            submission.feedback_status = 'feedback_outdated'

        if submission.manual_status == 'manually_graded':
            submission.manually_graded = 'being_edited'


        self.session.commit()
        self.write_json(submission)

    # TODO: not used, remove?
    def get_extra_credit(self, gradebook):
        extra_credit = 0
        for notebook in gradebook.notebooks.values():
            for grades in notebook.grades:
                extra_credit += grades.extra_credit if \
                    grades.extra_credit is not None else 0
        self.log.info("Extra credit is " + str(extra_credit))
        return extra_credit


@register_handler(
    path=r'\/api\/lectures\/(?P<lecture_id>\d*)\/assignments\/' +
         r'(?P<assignment_id>\d*)\/submissions\/(?P<submission_id>\d*)\/edit\/?',
    version_specifier=VersionSpecifier.ALL,
)
class SubmissionEditHandler(GraderBaseHandler):

    @authorize([Scope.tutor, Scope.instructor])
    async def put(self, lecture_id: int, assignment_id: int,
                  submission_id: int):
        """Create or overwrites the repository which stores changes of
        submissions files
        :param lecture_id: lecture id
        :param assignment_id: assignment id
        :param submission_id: submission id
        :return:
        """
        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)
        self.validate_parameters()

        submission = self.get_submission(lecture_id, assignment_id,
                                         submission_id)
        assignment = submission.assignment
        lecture = assignment.lecture

        # Path to repository which will store edited submission files
        git_repo_path = os.path.join(
            self.gitbase,
            lecture.code,
            str(assignment.id),
            "edit",
            str(submission_id),
        )

        # Path to repository of student which contains the submitted files
        submission_repo_path = os.path.join(
            self.gitbase,
            lecture.code,
            str(assignment.id),
            assignment.type,
            submission.username
        )

        if os.path.exists(git_repo_path):
            shutil.rmtree(git_repo_path)

        # Creating bare repository
        if not os.path.exists(git_repo_path):
            os.makedirs(git_repo_path, exist_ok=True)

        await self._run_command_async('git init --bare', git_repo_path)

        # Create temporary paths to copy the submission
        # files in the edit repository
        tmp_path = os.path.join(
            self.application.grader_service_dir,
            "tmp",
            lecture.code,
            str(assignment.id),
            "edit",
            str(submission.id),
        )

        tmp_input_path = os.path.join(
            tmp_path,
            "input"
        )

        tmp_output_path = os.path.join(
            tmp_path,
            "output"
        )

        if os.path.exists(tmp_path):
            shutil.rmtree(tmp_path, ignore_errors=True)

        os.makedirs(tmp_input_path, exist_ok=True)

        # Init local repository
        command = "git init"
        await self._run_command_async(command, tmp_input_path)

        # Pull user repository
        command = f'git pull "{submission_repo_path}" main'
        await self._run_command_async(command, tmp_input_path)
        self.log.info("Successfully cloned repo")

        # Checkout to correct submission commit
        command = f"git checkout {submission.commit_hash}"
        await self._run_command_async(command, tmp_input_path)
        self.log.info(f"Now at commit {submission.commit_hash}")

        # Copy files to output directory
        shutil.copytree(tmp_input_path, tmp_output_path,
                        ignore=shutil.ignore_patterns(".git"))

        # Init local repository
        command = "git init"
        await self._run_command_async(command, tmp_output_path)

        # Add edit remote
        command = f"git remote add edit {git_repo_path}"
        await self._run_command_async(command, tmp_output_path)
        self.log.info("Successfully added edit remote")

        # Switch to main
        command = "git switch -c main"
        await self._run_command_async(command, tmp_output_path)
        self.log.info("Successfully switched to branch main")

        # Add files to staging
        command = "git add -A"
        await self._run_command_async(command, tmp_output_path)
        self.log.info("Successfully added files to staging")

        # Commit Files
        command = 'git commit -m "Initial commit" '
        await self._run_command_async(command, tmp_output_path)
        self.log.info("Successfully commited files")

        # Push copied files
        command = "git push edit main"
        await self._run_command_async(command, tmp_output_path)
        self.log.info("Successfully pushed copied files")

        submission.edited = True
        self.session.commit()
        self.write_json(submission)


@register_handler(
    path=r"\/api\/lectures\/(?P<lecture_id>\d*)\/assignments\/(?P<assignment_id>\d*)\/submissions\/lti\/?",
    version_specifier=VersionSpecifier.ALL,
)
class LtiSyncHandler(GraderBaseHandler):
    cache_token = {"token": None, "ttl": datetime.datetime.now()}

    @authorize([Scope.instructor])
    async def get(self, lecture_id: int, assignment_id: int):
        # apply task synchronously without adding to queue
        results = lti_sync_task.delay(lecture_id, assignment_id, None, False)
        results = results.get(timeout=120)
        if results is None:
            raise HTTPError(HTTPStatus.NOT_FOUND, reason="Did not find syncable submissions.")
        else:
            self.write_json(results)



@register_handler(
    path=r"\/api\/lectures\/(?P<lecture_id>\d*)\/assignments\/(?P<assignment_id>\d*)\/submissions\/count\/?"
)
class SubmissionCountHandler(GraderBaseHandler):
    """
    Tornado Handler class for http requests to /lectures/{lecture_id}/assignments/{assignment_id}/submissions/count.
    """

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int, assignment_id: int):
        """ Returns the count of submissions made by the student for an assignment, no matter if submission
         was deleted or not.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        """
        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)

        role = self.get_role(lecture_id)

        usersubmissions_count = self.session.query(Submission).filter(
            Submission.assignid == assignment_id,
            Submission.username == role.username,
        ).count()

        self.write_json({"submission_count": usersubmissions_count})
        self.session.close()


import datetime
import json
import os
import time
from urllib.parse import urlparse

import jwt
from traitlets import Callable, Dict, Unicode, Union, Bool
from traitlets.config import SingletonConfigurable
from http import HTTPStatus
from tornado.escape import url_escape, json_decode
from tornado.httpclient import AsyncHTTPClient, HTTPClientError, HTTPRequest
from tornado.web import HTTPError

def default_lti_username_convert(username: str) -> str:
    return username

def default_enable_lti(lecture, assignment, submissions):
    return False

def default_enable_sync_on_feedback(lecture, assignment, submissions):
    return False

class LTISyncGrades(SingletonConfigurable):
    enabled = Union(
        [Bool(False),
         Callable(default_enable_lti)],
        allow_none=True,
        config=True,
        help="""
        Determines if the LTI Sync Grades plugin should be used, defaults to False.
        Is either a bool value or
        a function with the params (lecture, assignment, submissions) returning a bool.
        """,
    )
    
    sync_on_feedback = Union(
        [Bool(False),
         Callable(default_enable_sync_on_feedback)],
        allow_none=True,
        config=True,
        help="""
        Determines if submissions should be automatically synchronised, on feedback generation, defaults to False.
        Only synchronises scores when LTISyncGrades.enabled is True.
        Is either a bool value or
        a function with the params (lecture, assignment, submissions) returning a bool.
        """,
    )
    client_id = Unicode(None, config=True, allow_none=True)
    token_url = Unicode(None, config=True, allow_none=True)
    # function used to change the hub username to the lti sourcedid value
    help_msg = "Converts the grader service username to the lti sourced id."
    username_convert = Callable(default_value=default_lti_username_convert,
                                         config=True,
                                         allow_none=True,
                                         help=help_msg)
    username_match = Unicode("user_id",
                             config=True,
                             allow_none=False,
                             help="Sets which membership container attribute should be matched against the submission username")
    token_private_key = Union(
        [Unicode(os.environ.get('LTI_PRIVATE_KEY', None)),
         Callable(None)],
        allow_none=True,
        config=True,
        help="""
        Private Key used to encrypt bearer token request
        """,
    )
    resolve_lti_urls = Callable(default_value=None,
                                         config=True,
                                         allow_none=True,
                                         help="Returns membership and lineitem URL needed for grade sync")

    # cache for lti token    
    cache_token = {"token": None, "ttl": datetime.datetime.now()}

    def check_if_lti_enabled(self, lecture, assignment, submissions, feedback_sync):
        if callable(self.enabled):
            enable_lti = self.enabled(lecture, assignment, submissions)
        else:
            enable_lti = self.enabled

        if enable_lti:
            if feedback_sync:
                if self.sync_on_feedback:
                    return True
                else:
                    return False
            else:
                return True
        else:
            return False

    async def start(self, lecture, assignment, submissions):

        # # Check if the LTI plugin should be used
        # if not self._check_if_lti_enabled(lecture, assignment, submissions, sync_on_feedback):
        #     self.log.info("Skipping LTI plugin as it is not enabled")
        #     return {"syncable_users": 0, "synced_user": 0}

        self.log.info("LTI: start grade sync")
        # if len(submissions) == 0:
        #     raise HTTPError(HTTPStatus.BAD_REQUEST, reason="No submissions to sync")

        # 1. request bearer token
        self.log.debug("LTI: request bearer token")
        stamp = datetime.datetime.now()
        if self.cache_token["token"] and self.cache_token["ttl"] > stamp - datetime.timedelta(
                minutes=50):
            token = self.cache_token["token"]
        else:
            token = await self.request_bearer_token()
            self.cache_token["token"] = token
            self.cache_token["ttl"] = datetime.datetime.now()

        # 2. resolve lti urls
        self.log.debug("LTI: resolve lti url")
        try:
            lti_urls = self.resolve_lti_urls(lecture, assignment, submissions)
            self.log.debug(f"LTI membership and lineitems URL: {lti_urls}")
            lineitems_url = lti_urls["lineitems_url"]
            membership_url = lti_urls["membership_url"]
        except Exception as e:
            self.log.error(e)
            return
        # 3. get all members
        self.log.debug("LTI: request all members of lti course")
        httpclient = AsyncHTTPClient()
        try:
            response = await httpclient.fetch(HTTPRequest(url=membership_url, method="GET",
                                                          headers={
                                                            "Authorization": "Bearer " + self.cache_token["token"],
                                                            "Accept": "application/vnd.ims.lti-nrps.v2.membershipcontainer+json"}))
        except HTTPClientError as e:
            self.log.error(e.response)
            raise HTTPError(e.code, reason="Unable to get users of course:" + e.response.reason)
        members = json_decode(response.body)["members"]

        # 4. match usernames of submissions to lti memberships
        # and generate for each submission a request body -> grades list
        self.log.debug("LTI: match grader usernames with lti identifier")
        grades = []
        syncable_user_count = 0
        for submission in submissions:
            for member in members:
                try:
                    user_match = member[self.username_match]
                except KeyError:
                    self.log.error(f"LTI Error: Given username_match key '{self.username_match}' does not exist in LTI membership user")
                    continue
                if user_match == self.username_convert(submission["username"]):
                    syncable_user_count += 1
                    grades.append(self.build_grade_publish_body(member["user_id"], submission["score"],
                                                           float(assignment["points"])))
        self.log.info(f"LTI: matched {syncable_user_count} users")
        # 6. get all lineitems
        self.log.debug("LTI: resolve lti url")
        try:
            response = await httpclient.fetch(HTTPRequest(url=lineitems_url, method="GET",
                                                          headers={
                                                            "Authorization": "Bearer " + self.cache_token["token"],
                                                            "Accept": "application/vnd.ims.lis.v2.lineitemcontainer+json"}))
        except HTTPClientError as e:
            self.log.error(e.response)
            raise HTTPError(e.code, reason="Unable to get lineitems of course:" + e.response.reason)
        lineitems = json_decode(response.body)
        self.log.debug(f"LTI found lineitems: {lineitems}")

        # 7. check if a lineitem with assignment name exists
        lineitem = None
        for item in lineitems:
            if item["label"] == assignment["name"]:
                # lineitem found
                self.log.debug(f"LTI found lineitem: {item}")
                lineitem = item
                break

        # 8. if not create a lineitem with the assignment name
        if lineitem is None:
            lineitem_body = {"scoreMaximum": float(assignment["points"]), "label": assignment["name"],
                             "resourceId": assignment["id"],
                             "tag": "grade", "startDateTime": str(datetime.datetime.now()),
                             "endDateTime": str(datetime.date.today() + datetime.timedelta(days=1, hours=1))}
            try:
                response = await httpclient.fetch(HTTPRequest(url=lineitems_url, method="POST", body=json.dumps(lineitem_body),
                                                          headers={
                                                            "Authorization": "Bearer " + self.cache_token["token"],
                                                            "Content-Type": "application/vnd.ims.lis.v2.lineitem+json"}))
            except HTTPClientError as e:
                self.log.error(e.response)
                raise HTTPError(e.code, reason="Unable to create new lineitem in course:" + e.response.reason)
        # due to different "interpretations" of the ims lti standard,
        # the response is sometimes a list containing the lineitem or
        # just the lineitem json
            try:
                lineitem_response = json_decode(response.body)
            except Exception as e:
                self.log.error("LTI: could not decode lineitem request response")
                raise e
            if isinstance(lineitem_response, list):
                lineitem = lineitem_response[0]
            elif isinstance(lineitem_response, dict):
                lineitem = lineitem_response
            else:
                self.log.error("LTI: lineitem request response does not match dict or list")
                raise HTTPError(HTTPStatus.UNPROCESSABLE_ENTITY, "lineitem request response does not match dict or list")

        # 9. push grades to lineitem
        url_parsed = urlparse(lineitem["id"])
        lineitem = url_parsed._replace(path=url_parsed.path + "/scores").geturl()
        self.log.debug("LTI: start sending grades to LTI course")
        synced_user = 0
        for grade in grades:
            try:
                response = await httpclient.fetch(HTTPRequest(url=lineitem, method="POST", body=json.dumps(grade),
                                                          headers={
                                                            "Authorization": "Bearer " + self.cache_token["token"],
                                                            "Content-Type": "application/vnd.ims.lis.v1.score+json"}))
                synced_user += 1
            except HTTPClientError as e:
                self.log.error(e.response)
        self.log.info("LTI Grade Sync finished successfully")
        return {"syncable_users": syncable_user_count, "synced_user": synced_user}


    def build_grade_publish_body(self, uid: str, score: float, max_score: float):
        return {
            "timestamp": str(datetime.datetime.now().astimezone().replace(microsecond=0).isoformat()),
            "scoreGiven": score,
            "comment": "Automatically synced",
            "scoreMaximum": max_score,
            "activityProgress": "Submitted",
            "gradingProgress": "FullyGraded",
            "userId": uid
        }

    async def request_bearer_token(self):
        # get config variables
        if self.client_id is None:
            raise HTTPError(HTTPStatus.NOT_FOUND,
                            reason="Unable to request bearer token: client_id is not set in grader config")
        if self.token_url is None:
            raise HTTPError(HTTPStatus.NOT_FOUND,
                            reason="Unable to request bearer token: token_url is not set in grader config")

        private_key = self.token_private_key
        if private_key is None:
            raise HTTPError(HTTPStatus.NOT_FOUND,
                            reason="Unable to request bearer token: token_private_key is not set in grader config")
        if callable(private_key):
            private_key = private_key()

        payload = {"iss": "grader-service", "sub": self.client_id, "aud": [self.token_url],
                   "ist": str(int(time.time())), "exp": str(int(time.time()) + 60),
                   "jti": str(int(time.time())) + "123"}
        try:
            encoded = jwt.encode(payload, private_key, algorithm="RS256")
        except Exception as e:
            raise HTTPError(HTTPStatus.UNPROCESSABLE_ENTITY, f"Unable to encode payload: {str(e)}")

        scopes = [
            "https://purl.imsglobal.org/spec/lti-ags/scope/score",
            "https://purl.imsglobal.org/spec/lti-ags/scope/lineitem",
            "https://purl.imsglobal.org/spec/lti-nrps/scope/contextmembership.readonly"
        ]
        scopes = url_escape(" ".join(scopes))
        data = f"grant_type=client_credentials&client_assertion_type=urn%3Aietf%3Aparams%3Aoauth%3Aclient-assertion" \
               f"-type%3Ajwt-bearer&client_assertion={encoded}&scope={scopes} "

        httpclient = AsyncHTTPClient()
        try:
            response = await httpclient.fetch(HTTPRequest(url=self.token_url, method="POST", body=data,
                                                          headers={
                                                              "Content-Type": "application/x-www-form-urlencoded"}))
        except HTTPClientError as e:
            self.log.error(e.response)
            raise HTTPError(e.code, reason="Unable to request token:" + e.response.reason)
        return json_decode(response.body)["access_token"]


# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from grader_service.orm.lecture import Lecture
from grader_service.orm.user import User
from grader_service.orm.takepart import Role
from grader_service.orm.submission import Submission
from grader_service.orm.assignment import Assignment
from grader_service.orm.base import Base
from grader_service.orm.group import Group
from grader_service.orm.oauthcode import OAuthCode
from grader_service.orm.oauthclient import OAuthClient
from grader_service.orm.api_token import APIToken

__all__ = ['Lecture', 'User', 'Role', 'Submission',
           'Assignment', 'Base', 'Group']



from calendar import timegm
from datetime import datetime, timezone
from typing import Any, Dict, Iterable

import jwt
from traitlets import Int
from traitlets.config import LoggingConfigurable

from .constants import (
    LTI13_AUTH_RESPONSE_ARGS,
    LTI13_DEEP_LINKING_REQUIRED_CLAIMS,
    LTI13_GENERAL_REQUIRED_CLAIMS,
    LTI13_INIT_LOGIN_REQUEST_ARGS,
    LTI13_RESOURCE_LINK_REQUEST_REQUIRED_CLAIMS,
)
from .error import (
    IncorrectValueError,
    InvalidAudienceError,
    MissingRequiredArgumentError,
    TokenError,
)


def now() -> int:
    return timegm(datetime.now(tz=timezone.utc).utctimetuple())


class LTI13LaunchValidator(LoggingConfigurable):
    """
    Allows JupyterHub to verify LTI 1.3 compatible requests as a tool (known as a tool
    provider with LTI 1.1).
    """

    time_leeway = Int(
        0,
        config=True,
        help="""
        A time margin in seconds for the JWT expiration checks.

        It will be added as a grace period to mitigate clock drift issues.
        """,
    )

    max_age = Int(
        600,
        config=True,
        help="""
        Maximum age of an id_token to be accepted.
        """,
    )

    def validate_login_request(self, args: Dict[str, Any]) -> None:
        """
        Validates the initial authentication request and ensures the required
        claims are included in said request.

        Args:
          args: dictionary that represents keys/values sent in authentication request

        Raises:
          MissingRequiredArgumentError if an argument is missing
        """
        self._validate_required_args(args, LTI13_INIT_LOGIN_REQUEST_ARGS)

    def validate_auth_response(self, args: Dict[str, Any]) -> None:
        """
        Validate the reponse from the authorization server and ensures that the required
        parameters are present.

        Args:
          args: dictionary that represents keys/values sent in authentication response

        Raises:
          MissingRequiredArgumentError if an required argument is missing or has no value
        """
        self._validate_required_args(args, LTI13_AUTH_RESPONSE_ARGS)

    def _check_arg_not_missing(self, args: Dict[str, Any], required: Iterable[str]):
        for a in required:
            if a not in args:
                raise MissingRequiredArgumentError(
                    f"Required LTI 1.3 arg {a} not in request"
                )

    def _check_arg_not_empty(self, args: Dict[str, Any], required: Iterable[str]):
        for a in required:
            if not args.get(a):
                raise MissingRequiredArgumentError(
                    f"Required LTI 1.3 arg {a} needs a value"
                )

    def _validate_required_args(self, args: Dict[str, Any], required: Iterable[str]):
        self._check_arg_not_missing(args, required)
        self._check_arg_not_empty(args, required)

    def verify_and_decode_jwt(
        self, encoded_jwt, issuer, audience, jwks_endpoint, jwks_algorithms, **kwargs
    ):
        """
        Verify the JWT against the public keys provided in a JSON Web Key Set
        endpoint provided by the platform, and then return the payload in the
        jwt.
        """
        if not issuer:
            self.log.warning("No issuer identifyer configured")

        # reject unsigned id_token if encription is negotiated
        verification_options = kwargs.pop("options", {})
        verification_options["verify_signature"] = (
            verification_options.get("verify_signature", True) or jwks_endpoint != ""
        )

        try:
            if verification_options["verify_signature"]:
                jwks_client = jwt.PyJWKClient(jwks_endpoint)
                signing_key = jwks_client.get_signing_key_from_jwt(encoded_jwt)
                key = signing_key.key
            else:
                key = ""

            id_token = jwt.decode(
                encoded_jwt,
                key,
                algorithms=jwks_algorithms,
                audience=audience,
                issuer=issuer,
                options=verification_options,
                leeway=self.time_leeway,
                **kwargs,
            )
        except jwt.InvalidAudienceError as e:
            raise InvalidAudienceError(str(e))
        except jwt.PyJWTError as e:
            raise TokenError(str(e))

        return id_token

    def _check_message_type(self, id_token: Dict[str, Any]) -> None:
        """Validate message type value"""
        message_type = id_token.get(
            "https://purl.imsglobal.org/spec/lti/claim/message_type"
        )
        if (
            message_type
            != LTI13_RESOURCE_LINK_REQUEST_REQUIRED_CLAIMS[
                "https://purl.imsglobal.org/spec/lti/claim/message_type"
            ]
            and message_type
            != LTI13_DEEP_LINKING_REQUIRED_CLAIMS[
                "https://purl.imsglobal.org/spec/lti/claim/message_type"
            ]
        ):
            raise IncorrectValueError(
                f"Incorrect value {message_type} for message_type claim"
            )

    def _check_context_label(self, id_token: Dict[str, Any]) -> None:
        """Validate context label."""
        context_claim = id_token.get(
            "https://purl.imsglobal.org/spec/lti/claim/context"
        )
        context_label = context_claim.get("label") if context_claim else None
        if context_label == "":
            raise MissingRequiredArgumentError(
                "Claim https://purl.imsglobal.org/spec/lti/claim/context present, but context label is missing."
            )

    def _check_lti_version(self, id_token: Dict[str, Any]) -> None:
        """Check version claim."""
        lti_version = id_token.get("https://purl.imsglobal.org/spec/lti/claim/version")
        if lti_version != "1.3.0":
            raise IncorrectValueError(
                f"Incorrect value {lti_version} for version claim"
            )

    def _check_resource_link_id(self, id_token: Dict[str, Any]) -> None:
        """Check if resource_link claim has id set."""
        id = id_token.get(
            "https://purl.imsglobal.org/spec/lti/claim/resource_link", {}
        ).get("id")
        if not id:
            raise MissingRequiredArgumentError(
                "resource_link claim's id can't be empty"
            )

    def validate_azp_claim(
        self, id_token: Dict[str, Any], client_id: Iterable[str]
    ) -> None:
        """Check if azp claim is present and matches client_id."""
        aud = id_token["aud"]
        need_azp = isinstance(id_token["aud"], list) and len(aud) > 1
        if not need_azp:
            return
        azp = id_token.get("azp")
        if not azp:
            raise MissingRequiredArgumentError(
                "azp claim is missing although multiple values for aud are given."
            )
        if azp not in client_id:
            raise InvalidAudienceError("azp claim does not match client_id.")

    def _check_if_iat_is_too_old(self, id_token: Dict[str, Any]) -> None:
        """Raise TokenError if iat is too old."""
        iat = id_token["iat"]
        if iat < now() - self.max_age - self.time_leeway:
            raise TokenError("id_token issued too long ago.")

    def validate_id_token(self, id_token: Dict[str, Any]) -> None:
        """
        Validates that a LTI 1.3 launch request's decoded JWT has required
        claims (dictionary keys of id_token).

        The required claims are defined by the LTI 1.3 standard, see
        https://www.imsglobal.org/spec/lti/v1p3#required-message-claims.

        Args:
          id_token: decoded JWT payload of a launch request

        Raises:
          Subclass of ValidationError.
        """
        self._check_arg_not_missing(id_token, LTI13_GENERAL_REQUIRED_CLAIMS.keys())
        self._check_lti_version(id_token)
        self._check_context_label(id_token)
        self._check_message_type(id_token)
        self._check_if_iat_is_too_old(id_token)

        # message_type influences what claims are required
        message_type = id_token[
            "https://purl.imsglobal.org/spec/lti/claim/message_type"
        ]
        is_deep_linking = message_type == "LtiDeepLinkingRequest"

        if is_deep_linking:
            self._check_arg_not_missing(
                id_token, LTI13_DEEP_LINKING_REQUIRED_CLAIMS.keys()
            )
        else:
            self._check_arg_not_missing(
                id_token, LTI13_RESOURCE_LINK_REQUEST_REQUIRED_CLAIMS.keys()
            )
            self._check_resource_link_id(id_token)



from datetime import datetime, timezone

from sqlalchemy import Column, Integer, Unicode, ForeignKey, or_
from sqlalchemy.orm import relationship, joinedload

from grader_service.orm.base import Base
from grader_service.orm.json_util import JSONList


class OAuthCode(Base):
    __tablename__ = 'oauth_code'
    id = Column(Integer, primary_key=True, autoincrement=True)
    client_id = Column(
        Unicode(255),
        ForeignKey('oauth_client.identifier', ondelete='CASCADE')
    )
    client = relationship(
        "OAuthClient",
        back_populates="codes",
    )
    code = Column(Unicode(36))
    expires_at = Column(Integer)
    redirect_uri = Column(Unicode(1023))
    session_id = Column(Unicode(255))
    # state = Column(Unicode(1023))
    username = Column(Integer, ForeignKey('user.name', ondelete='CASCADE'))
    user = relationship(
        "User",
        back_populates="oauth_codes",
    )

    scopes = Column(JSONList, default=[])

    @staticmethod
    def now():
        return datetime.now(timezone.utc).timestamp()

    @classmethod
    def find(cls, db, code):
        return (
            db.query(cls)
            .filter(cls.code == code)
            .filter(or_(cls.expires_at is None, cls.expires_at >= cls.now()))
            .options(
                # load user with the code
                joinedload(cls.user, innerjoin=True),
            )
            .first()
        )

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}(id={self.id}, client_id={self.client_id!r})>"
        )





import logging
import os
import sys

from jupyter_core.application import JupyterApp
from traitlets.config.application import Application
from traitlets.traitlets import Unicode, validate, TraitError, Bool

from grader_service._version import __version__

base_converter_aliases = {
    "log-level": "Application.log_level",
    "i": "ConverterApp.input_directory",
    "o": "ConverterApp.output_directory",
    "p": "ConverterApp.file_pattern",
    "input_directory": "ConverterApp.input_directory",
    "output_directory": "ConverterApp.output_directory",
    "file_pattern": "ConverterApp.file_pattern",
    "copy_files": "ConverterApp.copy_files"
}
base_converter_flags = {
    "debug": (
        {"Application": {"log_level": "DEBUG"}},
        "set log level to DEBUG (maximize logging output)",
    ),
    "quiet": (
        {"Application": {"log_level": "CRITICAL"}},
        "set log level to CRITICAL (minimize logging output)",
    ),
}


class ConverterApp(Application):
    description = """Base app for converters
    """

    __version__ = __version__

    aliases = base_converter_aliases
    flags = base_converter_flags

    input_directory = Unicode(None, allow_none=False).tag(config=True)
    output_directory = Unicode(None, allow_none=False).tag(config=True)
    file_pattern = Unicode("*.ipynb", allow_none=False).tag(config=True)
    copy_files = Bool(False, allow_none=False).tag(config=True)

    def _log_level_default(self):
        return logging.INFO

    @validate("input_directory", "output_directory")
    def _dir_exits(self, proposal) -> str:
        if os.path.isdir(proposal["value"]):
            return proposal["value"]
        else:
            self.log.error(f'The path {proposal.value} of {proposal.trait.name} is not a directory')
            raise TraitError(f'The path {proposal.value} of {proposal.trait.name} is not a directory')
        
    @validate("config_path")
    def _config_path_exists(self, proposal) -> str:
        if os.path.isfile(proposal["value"]):
            return proposal["value"]
        else:
            self.log.error(f'The path {proposal.value} of {proposal.trait.name} is not a valid config file')
            raise TraitError(f'The path {proposal.value} of {proposal.trait.name} is not a valid config file')

    def fail(self, msg, *args):
        """Log the error msg using self.log.error and exit using sys.exit(1)."""
        self.log.error(msg, *args)
        sys.exit(1)



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
import asyncio
import base64
from contextlib import contextmanager
import re
import shlex
import subprocess
import datetime
import functools
import json
import os
import shutil
import sys
import time
import traceback
import uuid
from _decimal import Decimal
from http import HTTPStatus
from pathlib import Path
from typing import Any, Awaitable, Callable, Iterator, List, Optional, Union
from urllib.parse import urlparse, parse_qsl
from grader_service._version import __version__

from sqlalchemy.exc import SQLAlchemyError
from tornado.httputil import url_concat
from traitlets import Type, Integer, TraitType, Unicode
from traitlets import List as ListTrait
from traitlets.config import SingletonConfigurable

from grader_service.api.models.base_model import Model
from grader_service.api.models.error_message import ErrorMessage
from grader_service.utils import maybe_future, url_path_join, get_browser_protocol, utcnow
from grader_service.autograding.local_grader import LocalAutogradeExecutor
from grader_service.orm import Group, Assignment, Submission, APIToken
from grader_service.orm.base import Serializable, DeleteState
from grader_service.orm.lecture import Lecture
from grader_service.orm.takepart import Role, Scope
from grader_service.orm.user import User
from grader_service.registry import VersionSpecifier, register_handler
from grader_service.server import GraderServer
from sqlalchemy.orm.exc import MultipleResultsFound, NoResultFound
from tornado import httputil, web
from tornado.escape import json_decode
from tornado.web import HTTPError
from sqlalchemy.orm.session import Session

SESSION_COOKIE_NAME = 'grader-session-id'

auth_header_pat = re.compile(r'^(token|bearer|basic)\s+([^\s]+)$', flags=re.IGNORECASE)

def check_authorization(self: "GraderBaseHandler", scopes: list[Scope], lecture_id: Union[int, None]) -> bool:
    if (("/permissions" in self.request.path)
            or ("/config" in self.request.path)):
        return True
    if (
            lecture_id is None
            and "/lectures" in self.request.path
            and self.request.method == "POST"
    ):
        # lecture name and semester is in post body
        try:
            data = json_decode(self.request.body)
            lecture_id = (
                self.session.query(Lecture)
                .filter(Lecture.code == data["code"])
                .one()
                .id
            )
        except MultipleResultsFound:
            raise HTTPError(403)
        except NoResultFound:
            raise HTTPError(404, reason="Lecture not found")
        except json.decoder.JSONDecodeError:
            raise HTTPError(403)
    elif (
            lecture_id is None
            and "/lectures" in self.request.path
            and self.request.method == "GET"
    ):
        return True

    role = self.session.get(Role, (self.user.name, lecture_id))
    if (role is None) or (role.role not in scopes):
        msg = f"User {self.user.name} tried to access "
        msg += f"{self.request.path} with insufficient privileges"
        self.log.warning(msg)
        raise HTTPError(403)
    return True


def authorize(scopes: list[Scope]):
    """Checks if user is authorized.
    :param scopes: the user's roles
    :return: wrapper function
    """
    if not set(scopes).issubset({Scope.student, Scope.tutor,
                                 Scope.instructor}):
        return ValueError("Invalid scopes")

    def wrapper(handler_method):
        @functools.wraps(handler_method)
        async def request_handler_wrapper(self: "GraderBaseHandler", *args,
                                          **kwargs):
            lecture_id = self.path_kwargs.get("lecture_id", None)
            check_authorization(self, scopes, lecture_id)  # raises appropriate HTTPError if not authorized
            return await handler_method(self, *args, **kwargs)

        return request_handler_wrapper

    return wrapper

class BaseHandler(web.RequestHandler):
    """Base class of all handler classes

    Implements validation and request functions"""

    def data_received(self, chunk: bytes) -> Optional[Awaitable[None]]:
        pass

    def __init__(
            self,
            application: GraderServer,
            request: httputil.HTTPServerRequest,
            **kwargs: Any,
    ) -> None:
        super().__init__(application, request, **kwargs)
        # add type hint for application
        self._accept_cookie_auth = True
        self._accept_token_auth = True

        self.application: GraderServer = self.application
        self.authenticator = self.application.authenticator
        self.log = self.application.log

    async def prepare(self) -> Optional[Awaitable[None]]:
        #strip trailing slash
        self.request.path = self.request.path.rstrip("/")
        
        #start session
        self.session: Session = self.application.session_maker()
        
        #authenticate
        try:
            await self.get_current_user()

            # if user is not authenticated and is not actively trying to authenticate
            if not self.current_user and self.request.path not in [
                self.settings["login_url"],
                self.application.base_url.rstrip('/'),
                url_path_join(self.application.base_url, "/health"),
                url_path_join(self.application.base_url, "/api/oauth2/token"),
                url_path_join(self.application.base_url, "/oauth_callback"),
                url_path_join(self.application.base_url, "/lti13/oauth_callback")
                ]:
                # require git to authenticate with token -> otherwise return 401 code
                if self.request.path.startswith(url_path_join(self.application.base_url, "/git")):
                    raise HTTPError(401, reason="Git: authenticate request")
                
                # send to login page if ui page request
                if self.request.path in [url_path_join(self.application.base_url, "/api/oauth2/authorize")] or self.request.path.startswith(url_path_join(self.application.base_url, "/ui")):
                    url = url_concat(self.settings["login_url"], dict(next=self.request.uri))
                    self.redirect(url)
                    return
                    
                if self.request.headers.get("Authorization") is None:
                    raise HTTPError(401, reason="No API token in auth header")
                    
                # do not redirect to login page if we hit api endpoints
                raise HTTPError(401, reason="API Token is invalid or expired.")
                
                
                
        except Exception as e:
            # ensure get_current_user is never called again for this handler,
            # since it failed
            self._grader_user = None
            self.log.exception("Failed to get current user")
            if isinstance(e, SQLAlchemyError):
                self.log.error("Rolling back session due to database error")
                self.session.rollback()
            if isinstance(e, HTTPError) and e.status_code == 401:
                raise e
        await maybe_future(super().prepare())
        return

    @property
    def oauth_provider(self):
        return self.application.oauth_provider

    @property
    def csp_report_uri(self):
        return self.settings.get(
            'csp_report_uri',
            url_path_join(self.application.base_url, 'security/csp-report')
        )

    @property
    def content_security_policy(self):
        """The default Content-Security-Policy header

        Can be overridden by defining Content-Security-Policy in settings['headers']
        """
        return '; '.join(
            ["frame-ancestors 'self'", "report-uri " + self.csp_report_uri]
        )

    def _set_cookie(self, key, value, encrypted=True, **overrides):
        """Setting any cookie should go through here

        if encrypted use tornado's set_secure_cookie,
        otherwise set plaintext cookies.
        """
        # tornado <4.2 have a bug that consider secure==True as soon as
        # 'secure' kwarg is passed to set_secure_cookie
        kwargs = {'httponly': True}
        public_url = self.settings.get("public_url")
        if public_url:
            if public_url.scheme == 'https':
                kwargs['secure'] = True
        else:
            if self.request.protocol == 'https':
                kwargs['secure'] = True

        kwargs.update(self.settings.get('cookie_options', {}))
        kwargs.update(overrides)

        if key.startswith("__Host-"):
            # __Host- cookies must be secure and on /
            kwargs["path"] = "/"
            kwargs["secure"] = True

        if encrypted:
            set_cookie = self.set_secure_cookie
        else:
            set_cookie = self.set_cookie

        self.log.debug("Setting cookie %s: %s", key, kwargs)
        set_cookie(key, value, **kwargs)

    def _set_user_cookie(self, user, server: "GraderServer"):
        self.log.debug("Setting cookie for %s: %s", user.name,
                       server.cookie_name)
        self._set_cookie(
            server.cookie_name, user.cookie_id, encrypted=True,
            path=server.base_url.rstrip('/')
        )

    def clear_login_cookies(self):
        kwargs = {}
        user = self.get_current_user_cookie()
        session_id = self.get_session_cookie()
        if session_id:
            # clear session id
            session_cookie_kwargs = {}
            session_cookie_kwargs.update(kwargs)

            self.clear_cookie(
                SESSION_COOKIE_NAME, path=self.application.base_url.rstrip('/'), **session_cookie_kwargs
            )

            if user:
                # user is logged in, clear any tokens associated with the current session
                # don't clear session tokens if not logged in,
                # because that could be a malicious logout request!
                count = 0
                for access_token in self.session.query(APIToken).filter_by(
                        username=user.name, session_id=session_id
                ):
                    self.session.delete(access_token)
                    count += 1
                if count:
                    self.log.debug("Deleted %s access tokens for %s", count, user.name)
                    self.session.commit()

        # clear hub cookie
        self.clear_cookie(self.application.cookie_name, path=self.application.base_url.rstrip('/'), **kwargs)

    def get_session_cookie(self):
        """Get the session id from a cookie

        Returns None if no session id is stored
        """
        return self.get_cookie(SESSION_COOKIE_NAME, None)

    def _user_for_cookie(self, cookie_name, cookie_value=None):
        """Get the User for a given cookie, if there is one"""
        cookie_id = self.get_secure_cookie(
            cookie_name, cookie_value, max_age_days=self.application.cookie_max_age_days
        )

        def clear():
            self.clear_cookie(cookie_name, path=self.application.base_url.rstrip('/'))

        if cookie_id is None:
            if self.get_cookie(cookie_name):
                self.log.warning("Invalid or expired cookie token")
                clear()
            return
        cookie_id = cookie_id.decode('utf8', 'replace')
        user = self.session.query(User).filter(
            User.cookie_id == cookie_id).first()
        # user = self._user_from_orm(u)
        if user is None:
            self.log.warning("Invalid cookie token")
            # have cookie, but it's not valid. Clear it and start over.
            clear()
            return
        # TODO: update user activity
        # if self._record_activity(user):
        #     self.session.commit()
        return user

    def _record_activity(self, obj, timestamp=None):
        """record activity on an ORM object

        If last_activity was more recent than self.activity_resolution seconds ago,
        do nothing to avoid unnecessarily frequent database commits.

        Args:
            obj: an ORM object with a last_activity attribute
            timestamp (datetime, optional): the timestamp of activity to register.
        Returns:
            recorded (bool): True if activity was recorded, False if not.
        """
        if timestamp is None:
            timestamp = utcnow(with_tz=False)
        resolution = self.settings.get("activity_resolution", 0)
        if not obj.last_activity or resolution == 0:
            self.log.debug("Recording first activity for %s", obj)
            obj.last_activity = timestamp
            return True
        if (timestamp - obj.last_activity).total_seconds() > resolution:
            # this debug line will happen just too often
            # uncomment to debug last_activity updates
            # self.log.debug("Recording activity for %s", obj)
            obj.last_activity = timestamp
            return True
        return False

    def get_auth_token(self):
        """Get the authorization token from Authorization header"""
        auth_header = self.request.headers.get('Authorization', '')
        match = auth_header_pat.match(auth_header)
        if not match:
            return None

        if match.group(1).lower() == "basic":
            auth_decoded = base64.b64decode(match.group(2)).decode('ascii')
            _, token = auth_decoded.split(":", 2)
            return token
        else:
            return match.group(2)

    @functools.lru_cache
    def get_token(self):
        """get token from authorization header"""
        token = self.get_auth_token()
        if token is None:
            return None
        orm_token = APIToken.find(self.session, token)
        return orm_token

    def get_current_user_token(self):
        """get_current_user from Authorization header token"""
        # record token activity
        orm_token = self.get_token()
        if orm_token is None:
            return None
        now = utcnow(with_tz=False)
        recorded = self._record_activity(orm_token, now)
        if recorded:
            self.session.commit()

        # record that we've been token-authenticated
        # XSRF checks are skipped when using token auth
        self._token_authenticated = True
        return orm_token.user

    def get_current_user_cookie(self):
        """get_current_user from a cookie token"""
        return self._user_for_cookie(self.application.cookie_name)

    async def refresh_auth(self, user, force=False):
        """Refresh user authentication info

        Calls `authenticator.refresh_user(user)`

        Called at most once per user per request.

        Args:
            user (User): the user whose auth info is to be refreshed
            force (bool): force a refresh instead of checking last refresh time
        Returns:
            user (User): the user having been refreshed,
                or None if the user must login again to refresh auth info.
        """
        refresh_age = self.authenticator.auth_refresh_age
        if not refresh_age:
            return user
        now = time.monotonic()
        if (
                not force
                and user._auth_refreshed
                and (now - user._auth_refreshed < refresh_age)
        ):
            # auth up-to-date
            return user

        # refresh a user at most once per request
        if not hasattr(self, '_refreshed_users'):
            self._refreshed_users = set()
        if user.name in self._refreshed_users:
            # already refreshed during this request
            return user
        self._refreshed_users.add(user.name)

        self.log.debug("Refreshing auth for %s", user.name)
        auth_info = await self.authenticator.refresh_user(user, self)

        if not auth_info:
            self.log.warning(
                "User %s has stale auth info. Login is required to refresh.", user.name
            )
            return

        user._auth_refreshed = now

        if auth_info == True:
            # refresh_user confirmed that it's up-to-date,
            # nothing to refresh
            return user

        # Ensure name field is set. It cannot be updated.
        auth_info['name'] = user.name

        if 'auth_state' not in auth_info:
            # refresh didn't specify auth_state,
            # so preserve previous value to avoid clearing it
            auth_info['auth_state'] = await user.get_auth_state()
        return await self.auth_to_user(auth_info, user)

    async def get_current_user(self):
        """get current username"""
        if not hasattr(self, '_grader_user'):
            user = None
            try:
                if self._accept_token_auth:
                    user = self.get_current_user_token()
                if user is None and self._accept_cookie_auth:
                    user = self.get_current_user_cookie()
                if user and isinstance(user, User):
                    user = await self.refresh_auth(user)
                self._grader_user = user
            except Exception:
                # don't let errors here raise more than once
                self._grader_user = None
                # but still raise, which will get handled in .prepare()
                raise
        return self._grader_user

    def on_finish(self):
        self.session.close()

    @property
    def current_user(self) -> User:
        """Override .current_user accessor from tornado

        Allows .get_current_user to be async.
        """
        if not hasattr(self, '_grader_user'):
            raise RuntimeError("Must call async get_current_user first!")
        return self._grader_user

    @property
    def user(self) -> User:
        return self.current_user

    def set_session_cookie(self):
        """Set a new session id cookie

        new session id is returned

        Session id cookie is *not* encrypted,
        so other services on this domain can read it.
        """
        session_id = uuid.uuid4().hex
        self._set_cookie(
            SESSION_COOKIE_NAME, session_id, encrypted=False,
            path=self.application.base_url.rstrip('/')
        )
        return session_id

    def set_grader_cookie(self, user):
        """set the login cookie for the Hub"""
        self._set_user_cookie(user, self.application)

    def set_login_cookie(self, user):
        """Set login cookies for the Hub and single-user server."""

        if not self.get_session_cookie():
            self.set_session_cookie()

        # create and set a new cookie for the hub
        cookie_user = self.get_current_user_cookie()
        if cookie_user is None or cookie_user.name != user.name:
            if cookie_user:
                self.log.info(f"User {cookie_user.name} is logging in as {user.name}")
            self.set_grader_cookie(user)

        # make sure xsrf cookie is updated
        # this avoids needing a second request to set the right xsrf cookie
        self._grader_user = user
        # _set_xsrf_cookie(
        #     self, self._xsrf_token_id, cookie_path=self.application.base_url, authenticated=True
        # )

    def authenticate(self, data):
        return maybe_future(
            self.authenticator.get_authenticated_user(self, data))

    async def auth_to_user(self, authenticated, user=None):
        """Persist data from .authenticate() or .refresh_user() to the User database

        Args:
            authenticated(dict): return data from .authenticate or .refresh_user
            user(User, optional): the User object to refresh, if refreshing
        Return:
            user(User): the constructed User object
        """
        if isinstance(authenticated, str):
            authenticated = {'name': authenticated}
        username = authenticated['name']
        auth_state = authenticated.get('auth_state')
        admin = authenticated.get('admin')
        refreshing = user is not None

        if user and username != user.name:
            raise ValueError(
                f"Username doesn't match! {username} != {user.name}")

        user_model = self.session.query(User).get(username)
        if user_model is None:
            self.log.info(f'User {username} does not exist and will be created.')
            user_model = User()
            user_model.name = username
            self.session.add(user_model)
            self.session.commit()

        # apply authenticator-managed groups
        if self.authenticator.manage_groups:
            if "groups" not in authenticated:
                # to use manage_groups, group membership must always be specified
                # Authenticators that don't support this feature will omit it,
                # which should fail here rather than silently not implement the requested behavior
                auth_cls = self.authenticator.__class__.__name__
                raise ValueError(
                    f"Authenticator.manage_groups is enabled, but auth_model for {username} specifies no groups."
                    f" Does {auth_cls} support manage_groups=True?"
                )
            group_names = authenticated["groups"]
            if group_names is not None:
                user.sync_groups(group_names)
        # apply authenticator-managed roles
        if self.authenticator.manage_roles:
            auth_roles = authenticated.get("roles")
            if auth_roles is not None:
                user.sync_roles(auth_roles)

        # always set auth_state and commit,
        # because there could be key-rotation or clearing of previous values
        # going on.
        if not self.authenticator.enable_auth_state:
            # auth_state is not enabled. Force None.
            auth_state = None

        await user_model.save_auth_state(auth_state)
        return user_model

    async def login_user(self, data=None):
        """Login a user"""
        # auth_timer = self.statsd.timer('login.authenticate').start()
        authenticated = await self.authenticate(data)
        # auth_timer.stop(send=False)

        if authenticated:
            user = await self.auth_to_user(authenticated)
            self.set_login_cookie(user)

            self.log.info("User logged in: %s", user.name)
            user._auth_refreshed = time.monotonic()
            return user
        else:
            self.log.warning(
                "Failed login for %s",
                (data or {}).get('username', 'unknown user')
            )

    def get_template(self, name, sync=False):
        """
        Return the jinja template object for a given name

        If sync is True, we return a Template that is compiled without async support.
        Only those can be used in synchronous code.

        If sync is False, we return a Template that is compiled with async support
        """
        if sync:
            return self.application.jinja_env_sync.get_template(name)
        else:
            return self.application.jinja_env.get_template(name)

    def render_template(self, name, sync=False, **ns):
        """
        Render jinja2 template

        If sync is set to True, we render the template & return a string
        If sync is set to False, we return an awaitable
        """
        template_ns = {}
        template_ns.update(self.template_namespace)
        template_ns["xsrf_token"] = self.xsrf_token.decode("ascii")
        template_ns.update(ns)
        template = self.get_template(name, sync)
        if sync:
            return template.render(**template_ns)
        else:
            return template.render_async(**template_ns)

    @property
    def parsed_scopes(self) -> set:
        # TODO: if user is admin, the scopes should contain "admin-ui" for login.html template
        scopes = set()
        return scopes

    @property
    def template_namespace(self):
        user = self.current_user
        base_url = os.path.join(self.application.base_url, "")  # make sure "/" is at the end
        ns = dict(
            base_url=base_url,
            prefix=base_url,
            user=user,
            login_url=self.settings['login_url'],
            login_service=self.authenticator.login_service,
            logout_url=self.settings['logout_url'],
            static_url=self.static_url,
            version_hash='',
            parsed_scopes=self.parsed_scopes,
            xsrf=self.xsrf_token.decode('ascii'),
        )
        if self.application.template_vars:
            for key, value in self.application.template_vars.items():
                if callable(value):
                    value = value(user)
                ns[key] = value
        return ns

    def _validate_next_url(self, next_url):
        """Validate next_url handling

        protects against external redirects, etc.

        Returns empty string if next_url is not considered safe,
        resulting in same behavior as if next_url is not specified.
        """
        # protect against some browsers' buggy handling of backslash as slash
        next_url = next_url.replace('\\', '%5C')
        public_url = self.settings.get("public_url")
        if public_url:
            proto = public_url.scheme
            host = public_url.netloc
        else:
            # guess from request
            proto = get_browser_protocol(self.request)
            host = self.request.host

        if next_url.startswith("///"):
            # strip more than 2 leading // down to 2
            # because urlparse treats that as empty netloc,
            # whereas browsers treat more than two leading // the same as //,
            # so netloc is the first non-/ bit
            next_url = "//" + next_url.lstrip("/")
        parsed_next_url = urlparse(next_url)

        if (next_url + '/').startswith(
                (
                        f'{proto}://{host}/',
                        f'//{host}/',
                )
        ):
            # treat absolute URLs for our host as absolute paths:
            # below, redirects that aren't strictly paths are rejected
            next_url = parsed_next_url.path
            if parsed_next_url.query:
                next_url = next_url + '?' + parsed_next_url.query
            if parsed_next_url.fragment:
                next_url = next_url + '#' + parsed_next_url.fragment
            parsed_next_url = urlparse(next_url)

        # if it still has host info, it didn't match our above check for *this* host
        if next_url and (parsed_next_url.netloc or not next_url.startswith('/')):
            self.log.warning("Disallowing redirect outside JupyterHub: %r", next_url)
            next_url = ''

        return next_url

    def get_next_url(self, user=None, default=None):
        """Get the next_url for login redirect

        Default URL after login:

        - if redirect_to_server (default): send to user's own server
        - else: /hub/home
        """
        next_url = self.get_argument('next', default='')
        next_url = self._validate_next_url(next_url)

        # this is where we know if next_url is coming from ?next= param or we are using a default url
        if next_url:
            next_url_from_param = True
        else:
            next_url_from_param = False

        if not next_url:
            # custom default URL, usually passed because user landed on that page but was not logged in
            if default:
                next_url = default
            else:
                # As set in jupyterhub_config.py
                if callable(self.authenticator.login_redirect_url):
                    next_url = self.authenticator.login_redirect_url(self)
                else:
                    next_url = url_path_join(self.application.base_url, self.authenticator.login_redirect_url)

        if not next_url_from_param:
            # when a request made with ?next=... assume all the params have already been encoded
            # otherwise, preserve params from the current request across the redirect
            next_url = self.append_query_parameters(next_url, exclude=['next', '_xsrf'])
        return next_url

    def append_query_parameters(self, url, exclude=None):
        """Append the current request's query parameters to the given URL.

        Supports an extra optional parameter ``exclude`` that when provided must
        contain a list of parameters to be ignored, i.e. these parameters will
        not be added to the URL.

        This is important to avoid infinite loops with the next parameter being
        added over and over, for instance.

        The default value for ``exclude`` is an array with "next". This is useful
        as most use cases in JupyterHub (all?) won't want to include the next
        parameter twice (the next parameter is added elsewhere to the query
        parameters).

        :param str url: a URL
        :param list exclude: optional list of parameters to be ignored, defaults to
        a list with "next" (to avoid redirect-loops)
        :rtype (str)
        """
        if exclude is None:
            exclude = ['next']
        if self.request.query:
            query_string = [
                param
                for param in parse_qsl(self.request.query)
                if param[0] not in exclude
            ]
            if query_string:
                url = url_concat(url, query_string)
        return url


class GraderBaseHandler(BaseHandler):
        
    def validate_parameters(self, *args):
        if len(self.request.arguments) == 0:
            return
        unknown_args = set(self.request.query_arguments.keys()) - set(args)
        if len(unknown_args) != 0:
            raise HTTPError(400, reason=f"Unknown arguments: {unknown_args}")

    def get_role(self, lecture_id: int) -> Role:
        role = self.session.get(Role, (self.user.name, lecture_id))
        if role is None:
            raise HTTPError(403)
        return role

    def get_assignment(self, lecture_id: int,
                       assignment_id: int) -> Assignment:
        assignment: Assignment = self.session.get(Assignment, assignment_id)
        if ((assignment is None) or (assignment.deleted == DeleteState.deleted)
                or (int(assignment.lectid) != int(lecture_id))):
            msg = "Assignment with id " + str(assignment_id) + " was not found"
            raise HTTPError(HTTPStatus.NOT_FOUND,
                            reason=msg)
        return assignment

    def get_submission(self, lecture_id: int, assignment_id: int,
                       submission_id: int) -> Submission:
        submission = self.session.get(Submission, submission_id)
        if (
                (submission is None)
                or (submission.assignid != assignment_id)
                or (int(submission.assignment.lectid) != int(lecture_id))
                or (submission.deleted == DeleteState.deleted)
        ):
            msg = "Submission with id " + str(submission_id) + " was not found"
            raise HTTPError(HTTPStatus.NOT_FOUND,
                            reason=msg)
        return submission

    @property
    def gitbase(self):
        app: GraderServer = self.application
        return os.path.join(app.grader_service_dir, "git")

    def construct_git_dir(self, repo_type: str, lecture: Lecture,
                          assignment: Assignment,
                          group_name: Optional[str] = None,
                          submission: Optional[Submission] = None
                          ) -> Optional[str]:
        """Helper method for every handler that needs to access git
        directories which returns the path of the repository based on
        the inputs or None if the repo_type is not recognized."""
        # TODO: refactor
        assignment_path = os.path.abspath(
            os.path.join(self.gitbase, lecture.code, str(assignment.id))
        )
        allowed_types = set(["source", "release", "edit"])
        if (repo_type in allowed_types):
            path = os.path.join(assignment_path, repo_type)
            if repo_type == "edit":
                path = os.path.join(path, str(submission.id))
                self.log.info(path)
        elif repo_type in ["autograde", "feedback"]:
            type_path = os.path.join(assignment_path, repo_type,
                                     assignment.type)
            if assignment.type == "user":
                if repo_type == "autograde":
                    if ((submission is None)
                            or (self.get_role(lecture.id).role < Scope.tutor)):
                        raise HTTPError(403)
                    path = os.path.join(type_path, submission.username)
                else:
                    path = os.path.join(type_path, self.user.name)
            else:
                group = self.session.query(Group).get((self.user.name,
                                                       lecture.id))
                if group is None:
                    raise HTTPError(404, reason="User is not in a group")
                path = os.path.join(type_path, group.name)
        elif repo_type == "user":
            user_path = os.path.join(assignment_path, repo_type)
            path = os.path.join(user_path, self.user.name)
        elif repo_type == "group":
            if group_name is None:
                return None
            group = self.session.query(Group).get((group_name, lecture.id))
            if group is None:
                raise HTTPError(404, reason="User is not in a group")
            group_path = os.path.join(assignment_path, repo_type)
            path = os.path.join(group_path, group.name)
        else:
            return None

        return path

    @staticmethod
    def is_base_git_dir(path: str) -> bool:
        try:
            out = subprocess.run(["git", "rev-parse", "--is-bare-repository"],
                                 cwd=path, capture_output=True)
            is_git = ((out.returncode == 0)
                      and ("true" in out.stdout.decode("utf-8")))
        except FileNotFoundError:
            is_git = False
        return is_git

    def duplicate_release_repo(self, repo_path_release: str,
                               repo_path_user: str,
                               assignment: Assignment, message: str,
                               checkout_main: bool = False):
        tmp_path_base = Path(
            self.application.grader_service_dir,
            "tmp",
            assignment.lecture.code,
            str(assignment.id),
            str(self.user.name))

        # Deleting dir
        if os.path.exists(tmp_path_base):
            shutil.rmtree(tmp_path_base)

        os.makedirs(tmp_path_base, exist_ok=True)
        tmp_path_release = tmp_path_base.joinpath("release")
        tmp_path_user = tmp_path_base.joinpath(self.user.name)

        self.log.info(f"Duplicating release repository {repo_path_release}")
        self.log.info(f"Temporary path used for copying: {tmp_path_base}")

        try:
            self._run_command(f"git clone -b main '{repo_path_release}'",
                              cwd=tmp_path_base)
            if checkout_main:
                self._run_command(f"git clone '{repo_path_user}'",
                                  cwd=tmp_path_base)
                self._run_command("git checkout -b main", cwd=tmp_path_user)
            else:
                self._run_command(f"git clone -b main '{repo_path_user}'",
                                  cwd=tmp_path_base)

            msg = f"Copying repo from {tmp_path_release} to {tmp_path_user}"
            self.log.info(msg)
            ignore = shutil.ignore_patterns(".git", "__pycache__")
            if (sys.version_info.major == 3) and (sys.version_info.minor >= 8):
                shutil.copytree(tmp_path_release, tmp_path_user,
                                ignore=ignore, dirs_exist_ok=True)
            else:
                for item in os.listdir(tmp_path_release):
                    s = os.path.join(tmp_path_release, item)
                    d = os.path.join(tmp_path_user, item)
                    if os.path.isdir(s):
                        shutil.copytree(s, d, ignore=ignore)
                    else:
                        shutil.copy2(s, d)
            cmd = 'sh -c \'git add -A'
            cmd += f'&& git commit --allow-empty -m "{message}"\''
            self._run_command(cmd, tmp_path_user)
            self._run_command("git push -u origin main", tmp_path_user)
        finally:
            shutil.rmtree(tmp_path_base)

    def _run_command(self, command, cwd=None, capture_output=False):
        # TODO currently there a two run_command functions,
        #  because duplicate_release_repo does not work
        #  with the _run_command_async
        try:
            self.log.info(f"Running: {command}")
            ret = subprocess.run(shlex.split(command), check=True, cwd=cwd,
                                 capture_output=True)
            if capture_output:
                return str(ret.stdout, 'utf-8')
        except subprocess.CalledProcessError as e:
            self.log.error(e.stderr)
            raise HTTPError(500, reason="Subprocess Error")
        except FileNotFoundError as e:
            self.log.error(e)
            raise HTTPError(404, reason="File not found")

    async def _run_command_async(self, command, cwd=None):
        """Starts a sub process and runs a cmd command

        Args:
            command str: command that is getting run.
            cwd (str, optional): states where the command is getting run.
                                 Defaults to None.

        Raises:
            GitError: returns appropriate git error"""
        self.log.info(f"Running: {command}")
        ret = await asyncio.create_subprocess_shell(
            command,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            cwd=cwd)
        await ret.wait()

    def write_json(self, obj) -> None:
        self.set_header("Content-Type", "application/json")
        chunk = GraderBaseHandler._serialize(obj)
        self.write(json.dumps(chunk))

    @classmethod
    def _serialize(cls, obj: object):
        if isinstance(obj, list):
            return [cls._serialize(o) for o in obj]
        if isinstance(obj, dict):
            return {k: cls._serialize(v) for k, v in obj.items()}
        if isinstance(obj, tuple):
            return tuple(cls._serialize(o) for o in obj)
        if isinstance(obj, Serializable):
            return cls._serialize(obj.serialize())
        if isinstance(obj, (str, int, float, complex)) or obj is None:
            return obj
        if isinstance(obj, datetime.datetime):
            obj = obj.replace(tzinfo=datetime.timezone.utc)
            return str(obj)
        if isinstance(obj, Decimal):
            return float(obj)
        if isinstance(obj, Model):
            return cls._serialize(obj.to_dict())
        return None


def authenticated(
        method: Callable[..., Optional[Awaitable[None]]]
) -> Callable[..., Optional[Awaitable[None]]]:
    """Decorate methods with this to require that the user be logged in.

    If the user is not logged in `tornado.web.HTTPError`
    with code 403 will be raised.
    """

    @functools.wraps(method)
    def wrapper(  # type: ignore
            self: GraderBaseHandler, *args, **kwargs
    ) -> Optional[Awaitable[None]]:
        if not self.current_user:
            raise HTTPError(403)
        return method(self, *args, **kwargs)

    return wrapper


@register_handler(r"\/?", VersionSpecifier.NONE)
class VersionHandler(GraderBaseHandler):
    async def get(self):
        self.write(f"Version {__version__}")


@register_handler(r"\/?", VersionSpecifier.V1)
class VersionHandlerV1(GraderBaseHandler):
    async def get(self):
        self.write("Version 1.0")


class RequestHandlerConfig(SingletonConfigurable):
    """This class exists to not avoid all request handlers to inherit
    from traitlets.config.Configurable and making all requests super
    slow. If a request handler needs configurable values, they can be
    accessed from this object."""
    autograde_executor_class = Type(default_value=LocalAutogradeExecutor,
                                    # TODO: why does using
                                    # LocalAutogradeExecutor give
                                    # subclass error?
                                    klass=object,
                                    allow_none=False, config=True)

    # Git server file policy defaults
    git_max_file_size_mb = Integer(80, allow_none=False, config=True)
    git_max_file_count = Integer(512, allow_none=False, config=True)
    # empty list allows everything
    git_allowed_file_extensions = ListTrait(TraitType(Unicode),
                                            default_value=[],
                                            allow_none=False,
                                            config=True)



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import asyncio
import json
import os
import shutil
import inspect
from asyncio import Task, run

from kubernetes.client import (V1Pod, CoreV1Api, V1ObjectMeta,
                               V1PodStatus, ApiException)
from traitlets import Callable, Unicode, Integer, List, Dict
from traitlets.config import LoggingConfigurable
from urllib3.exceptions import MaxRetryError

from grader_service.autograding.kube.util import (make_pod,
                                                  get_current_namespace)
from grader_service.autograding.local_grader import (LocalAutogradeExecutor,
                                                     rm_error)
from kubernetes import config

from grader_service.orm import Lecture, Submission
from grader_service.orm import Assignment


class GraderPod(LoggingConfigurable):
    """
    Wrapper for a kubernetes pod that supports polling of the pod's status.
    """
    poll_interval = Integer(default_value=1000,
                            allow_none=False,
                            help="Time in ms to wait before "
                                 "status is polled again.").tag(config=True)

    def __init__(self, pod: V1Pod, api: CoreV1Api, **kwargs):
        super().__init__(**kwargs)
        self.pod = pod
        self._client = api
        self.loop = asyncio.get_event_loop()
        self._polling_task = self.loop.create_task(self._poll_status())

    def stop_polling(self) -> None:
        self._polling_task.cancel()

    def poll(self) -> str:
        return self.loop.run_until_complete(self.polling)

    @property
    def polling(self) -> Task:
        return self._polling_task

    @property
    def name(self) -> str:
        return self.pod.metadata.name

    @property
    def namespace(self) -> str:
        return self.pod.metadata.namespace

    # https://kubernetes.io/docs/concepts/workloads/pods/pod-lifecycle/#pod-phase
    async def _poll_status(self) -> str:
        meta: V1ObjectMeta = self.pod.metadata
        while True:
            status: V1PodStatus = \
                self._client.read_namespaced_pod_status(
                    name=meta.name,
                    namespace=meta.namespace).status

            if status.phase == "Succeeded" or status.phase == "Failed":
                return status.phase
            # continue for Running, Unknown and Pending
            await asyncio.sleep(self.poll_interval / 1000)


def _get_image_name(lecture: Lecture, assignment: Assignment = None) -> str:
    """
    Default implementation of the resolve_image_name method
    which return the lecture code followed by '_image'.
    All the functions have the lecture and assignment available as parameters.
    The function can either be sync or async.
    :param lecture: Lecture to build the image name.
    :param assignment: Assignment to build the image name.
    :return: The image name as a string.
    """
    return f"{lecture.code}_image"


class KubeAutogradeExecutor(LocalAutogradeExecutor):
    """
    Runs an autograde job in a kubernetes cluster as a pod.
    The cluster has to have a shared persistent
    volume claim that is mounted in the input
    and output directories so that both the service and
    the executor pods have access to the files.
    The service account of the grader service has to have
    permission to get, update, create and delete pods, pod status and pod logs.
    """

    image_config_path = Unicode(default_value=None,
                                allow_none=True).tag(config=True)
    resolve_image_name = Callable(default_value=_get_image_name,
                                  allow_none=False).tag(config=True)
    kube_context = Unicode(default_value=None, allow_none=True,
                           help="Kubernetes context to load config from. "

                                "If the context is None (default), "
                                "the incluster config "
                                "will be used.").tag(config=True)

    resolve_node_selector = Callable(default_value=lambda _: None, allow_none=False,
                                     help="""Function that takes a lecture as input and outputs a node selector for 
                                     this lecture. The returned value has to be a dict[str, str] or None and is directly 
                                     passed to the Kubernetes API.""").tag(config=True)

    volume = Dict(default_value={},
                  allow_none=False).tag(config=True)

    extra_volumes = List(default_value=[],
                         allow_none=False).tag(config=True)

    extra_volume_mounts = List(default_value=[],
                               allow_none=False).tag(config=True)

    convert_executable = Unicode("grader-convert",
                                 allow_none=False).tag(config=True)

    namespace = Unicode(default_value=None, allow_none=True,
                        help="Namespace to deploy grader pods into. "
                             "If changed, correct roles to Serviceaccount "
                             "need to be applied.").tag(config=True)
    uid = Integer(default_value=1000, allow_none=False,
                  help="The User ID for the grader container").tag(config=True)

    def __init__(self, grader_service_dir: str,
                 submission: Submission, **kwargs):
        super().__init__(grader_service_dir, submission, **kwargs)
        self.lecture = self.assignment.lecture

        if self.kube_context is None:
            self.log.info(f"Loading in-cluster config for kube executor "
                          f"of submission {self.submission.id}")
            config.load_incluster_config()
        else:
            self.log.info(
                f"Loading cluster config '{self.kube_context}' "
                f"for kube executor of submission {self.submission.id}")
            config.load_kube_config(context=self.kube_context)
        self.client = CoreV1Api()

        if self.namespace is None:
            self.log.info(f"Setting Namespace "
                          f"for submission {self.submission.id}")
            self.namespace = get_current_namespace()

    def get_image(self) -> str:
        """
        Returns the image name based on the lecture and assignment.
        If an image config file exists and has
        been specified it will first be queried for an image name.
        If the image name cannot be found in the
        config file or none has been specified
        the image name will be determined by the resolve_image_name function
        which takes the lecture
        and assignment as parameters and is specified in the config.
        The default implementation of this function is to
         return the lecture code followed by '_image'.
        :return: The image name as determined by this method.
        """
        cfg = {}
        if self.image_config_path is not None:
            with open(self.image_config_path, "r") as f:
                cfg = json.load(f)
        try:
            lecture_cfg = cfg[self.lecture.code]
            if isinstance(lecture_cfg, str):
                return lecture_cfg
            else:
                return lecture_cfg[self.assignment.name]
        except KeyError:
            if inspect.iscoroutinefunction(self.resolve_image_name):
                return run(self.resolve_image_name(self.lecture, self.assignment))
            else:
                return self.resolve_image_name(self.lecture, self.assignment)

    def start_pod(self) -> GraderPod:
        """
        Starts a pod in the default namespace
        with the commit hash as the name of the pod.
        The image is determined by the get_image method.
        :return:
        """
        # The output path will not exist in the pod
        command = [self.convert_executable, "autograde", "-i",
                   self.input_path, "-o", self.output_path,
                   "-p", "*.ipynb",
                   f"--copy_files={self.assignment.allow_files}",
                   "--log-level=INFO",
                   f"--ExecutePreprocessor.timeout={self.timeout_func(self.assignment.lecture)}"]

        volumes = [self.volume] + self.extra_volumes

        volume_mounts = [{"name": "data", "mountPath": self.input_path,
                          "subPath": self.relative_input_path +
                                     "/submission_" + str(self.submission.id)},
                         {"name": "data", "mountPath": self.output_path,
                          "subPath": self.relative_output_path +
                                     "/submission_" + str(self.submission.id)}]
        volume_mounts = volume_mounts + self.extra_volume_mounts

        pod = make_pod(
            name=self.submission.commit_hash,
            cmd=command,
            image=self.get_image(),
            image_pull_policy=None,
            working_dir="/",
            volumes=volumes,
            volume_mounts=volume_mounts,
            labels=None,
            annotations=None,
            node_selector=self.resolve_node_selector(self.lecture),
            tolerations=None,
            run_as_user=self.uid,
        )

        self.log.info(f"Starting pod {pod.metadata.name}"
                      f" with command: {command}")
        pod = self.client.create_namespaced_pod(namespace=self.namespace,
                                                body=pod)
        return GraderPod(pod, self.client, config=self.config)

    def _run(self):
        """
        Runs the autograding process in a kubernetes pod
        which has to have access to the files in the
        input and output directory through a persistent volume claim.
        :return: Coroutine
        """
        if os.path.exists(self.output_path):
            shutil.rmtree(self.output_path, onerror=rm_error)

        os.makedirs(self.output_path, exist_ok=True)

        self._write_gradebook(self._put_grades_in_assignment_properties())

        grader_pod = None
        try:
            grader_pod = self.start_pod()
            self.log.info(f"Started pod {grader_pod.name} in namespace "
                          f"{grader_pod.namespace}")
            status = grader_pod.poll()
            self.grading_logs = self._get_pod_logs(grader_pod)
            self.log.info("Pod logs:\n" + self.grading_logs)
            if status == "Succeeded":
                self.log.info("Pod has successfully completed execution!")
            else:
                self.log.info("Pod has failed execution:")
                self._delete_pod(grader_pod)
                raise RuntimeError("Pod has failed execution!")
            # cleanup
            self._delete_pod(grader_pod)
        except ApiException as e:
            error_message = json.loads(e.body)
            if error_message["reason"] != "AlreadyExists" \
                    and grader_pod is not None:
                try:
                    namespace = grader_pod.namespace
                    self.client.delete_namespaced_pod(name=grader_pod.name,
                                                      namespace=namespace)
                except ApiException:
                    pass
            self.log.error(f'{error_message["reason"]}: '
                           f'{error_message["message"]}')
            raise RuntimeError("Pod has failed execution!")
        except MaxRetryError:
            self.log.error("Kubernetes client could not connect to cluster! "
                           "Is it running and specified correctly?")
            raise RuntimeError("Pod has failed execution!")

    def _delete_pod(self, pod: GraderPod):
        """
        Deletes the pod from the cluster after successful or failed execution.
        :param pod: The pod to delete.
        :return: None
        """
        self.log.info(
            f"Deleting pod '{pod.name}' in namespace '{pod.namespace}' "
            f"after execution status {pod.polling.result()}")
        self.client.delete_namespaced_pod(name=pod.name,
                                          namespace=pod.namespace)

    def _get_pod_logs(self, pod: GraderPod) -> str:
        """
        Returns the logs of the pod that were output during execution.
        :param pod: The pod to retrieve the logs from.
        :return: The logs as a string.
        """
        api_response: str = \
            self.client.read_namespaced_pod_log(name=pod.name,
                                                namespace=pod.namespace)
        return api_response.strip()





from typing import Tuple

from nbconvert.exporters.exporter import ResourcesDict
from nbformat.notebooknode import NotebookNode

from grader_service.convert.gradebook.gradebook import Gradebook
from grader_service.convert import utils
from grader_service.convert.preprocessors.base import NbGraderPreprocessor


class SaveAutoGrades(NbGraderPreprocessor):
    """Preprocessor for saving out the autograder grades into a database"""

    def preprocess(
        self, nb: NotebookNode, resources: ResourcesDict
    ) -> Tuple[NotebookNode, ResourcesDict]:
        # pull information from the resources
        self.notebook_id = resources["unique_key"]
        self.json_path = resources["output_json_path"]
        self.gradebook = Gradebook(self.json_path)

        with self.gradebook:
            # process the cells
            nb, resources = super(SaveAutoGrades, self).preprocess(nb, resources)

        return nb, resources

    def _add_score(self, cell: NotebookNode, resources: ResourcesDict) -> None:
        """Graders can override the autograder grades, and may need to
        manually grade written solutions anyway. This function adds
        score information to the database if it doesn't exist. It does
        NOT override the 'score' field, as this is the manual score
        that might have been provided by a grader.

        """
        # these are the fields by which we will identify the score
        # information
        grade = self.gradebook.find_grade(
            cell.metadata["nbgrader"]["grade_id"], self.notebook_id
        )

        # determine what the grade is
        auto_score, _ = utils.determine_grade(cell, self.log)
        grade.auto_score = auto_score

        # if there was previously a manual grade, or if there is no autograder
        # score, then we should mark this as needing review
        if (grade.manual_score is not None) or (grade.auto_score is None):
            grade.needs_manual_grade = True
        else:
            grade.needs_manual_grade = False

        self.gradebook.add_grade(
            cell.metadata["nbgrader"]["grade_id"], self.notebook_id, grade
        )

    def _add_comment(self, cell: NotebookNode, resources: ResourcesDict) -> None:
        comment = self.gradebook.find_comment(
            cell.metadata["nbgrader"]["grade_id"], self.notebook_id
        )
        if cell.metadata.nbgrader.get("checksum", None) == utils.compute_checksum(
            cell
        ) and not utils.is_task(cell):
            comment.auto_comment = "No response."
        else:
            comment.auto_comment = None

        self.gradebook.add_comment(
            cell.metadata["nbgrader"]["grade_id"], self.notebook_id, comment
        )

    def preprocess_cell(
        self, cell: NotebookNode, resources: ResourcesDict, cell_index: int
    ) -> Tuple[NotebookNode, ResourcesDict]:
        # if it's a grade cell, the add a grade
        if utils.is_grade(cell):
            self._add_score(cell, resources)

        if utils.is_solution(cell):
            self._add_comment(cell, resources)

        if utils.is_task(cell):
            self._add_comment(cell, resources)

        return cell, resources



# LTI 1.3
# Login initiation request arguments
# https://www.imsglobal.org/spec/security/v1p0/#step-1-third-party-initiated-login
from typing import Any, Dict

LTI13_CUSTOM_CLAIM = "https://purl.imsglobal.org/spec/lti/claim/custom"

LTI13_INIT_LOGIN_REQUEST_ARGS = [
    "iss",
    "login_hint",
    "target_link_uri",
]

# Authentication request arguments
# https://www.imsglobal.org/spec/security/v1p0/#step-2-authentication-request
LTI13_AUTH_REQUEST_ARGS = [
    "client_id",
    "login_hint",
    "lti_message_hint",
    "nonce",
    "prompt",
    "redirect_uri",
    "response_mode",
    "response_type",
    "scope",
    "state",
]

LTI13_AUTH_RESPONSE_ARGS = [
    "id_token",
    "state",
]

# Required message claims
# http://www.imsglobal.org/spec/lti/v1p3/#required-message-claims
# https://openid.net/specs/openid-connect-core-1_0.html#IDToken
LTI13_GENERAL_REQUIRED_CLAIMS: Dict[str, Any] = {
    "https://purl.imsglobal.org/spec/lti/claim/version": "1.3.0",
    "https://purl.imsglobal.org/spec/lti/claim/deployment_id": "",
    "https://purl.imsglobal.org/spec/lti/claim/target_link_uri": "",
    "https://purl.imsglobal.org/spec/lti/claim/roles": "",
    "aud": "",
    "exp": None,
    "iat": None,
    "iss": "",
    "nonce": "",
}

# Required claims with LtiResourceLinkRequest login flows
LTI13_RESOURCE_LINK_REQUEST_REQUIRED_CLAIMS: Dict[str, Any] = {
    "https://purl.imsglobal.org/spec/lti/claim/message_type": "LtiResourceLinkRequest",
    "https://purl.imsglobal.org/spec/lti/claim/resource_link": {
        "id": "",
    },
    **LTI13_GENERAL_REQUIRED_CLAIMS,
}

# Required claims with LtiDeepLinkingRequest login flows
LTI13_DEEP_LINKING_REQUIRED_CLAIMS = {
    "https://purl.imsglobal.org/spec/lti/claim/message_type": "LtiDeepLinkingRequest",
    **LTI13_GENERAL_REQUIRED_CLAIMS,
}

# Optional message claims
# We don't need the role_scope_mentor claim and some platforms don't send this claim by default.
# http://www.imsglobal.org/spec/lti/v1p3/#optional-message-claims
LTI13_GENERAL_OPTIONAL_CLAIMS: Dict[str, Any] = {
    "https://purl.imsglobal.org/spec/lti/claim/context": {
        "id": "",
        "label": "",
        "title": "",
        "type": [
            "http://purl.imsglobal.org/vocab/lis/v2/course#CourseTemplate",
            "http://purl.imsglobal.org/vocab/lis/v2/course#CourseOffering",
            "http://purl.imsglobal.org/vocab/lis/v2/course#CourseSection",
            "http://purl.imsglobal.org/vocab/lis/v2/course#Group",
        ],
    },
    # user identity claims. sub (subject) is added to optional list to support anonymous
    # launches.
    "sub": "",
    "azp": "",
    "https://purl.imsglobal.org/spec/lti/claim/tool_platform": {
        "guid": "",
        "name": "",
        "version": "",
        "product_family_code": "",
        "validation_context": None,
        "errors": {"errors": {}},
    },
    "https://purl.imsglobal.org/spec/lti/claim/launch_presentation": {
        # one of frame, iframe, window
        "document_target": "",
        "height": "",
        "width": "",
        "return_url": "",
        "locale": "",
    },
}

# Optional resource link request claims
LTI13_RESOURCE_LINK_OPTIONAL_CLAIMS = {
    **LTI13_GENERAL_OPTIONAL_CLAIMS,
    "https://purl.imsglobal.org/spec/lti/claim/resource_link": {
        "description": "",
        "title": "",
    },
}

# Required claims for deep linking request claims
LTI13_DEEP_LINKING_REQUIRED_CLAIMS = {
    **LTI13_DEEP_LINKING_REQUIRED_CLAIMS,
    **LTI13_GENERAL_OPTIONAL_CLAIMS,
}

# Optional learning information system (LIS) claims
LTI13_LIS_CLAIMS = {
    "https://purl.imsglobal.org/spec/lti/claim/lis": {
        "course_offering_sourcedid": "",
        "course_section_sourcedid": "",
        "outcome_service_url": "",
        "person_sourcedid": "",
        "result_sourcedid": "",
    },
}

# Required and optional resource link claims
LTI13_RESOURCE_LINKS = {
    **LTI13_RESOURCE_LINK_REQUEST_REQUIRED_CLAIMS[
        "https://purl.imsglobal.org/spec/lti/claim/resource_link"
    ],
    **LTI13_RESOURCE_LINK_OPTIONAL_CLAIMS[
        "https://purl.imsglobal.org/spec/lti/claim/resource_link"
    ],
}

# Required claims for DeepLinkingRequest
LTI13_DEEP_LINKING_REQUIRED_CLAIMS = {
    **LTI13_DEEP_LINKING_REQUIRED_CLAIMS,
    **LTI13_GENERAL_OPTIONAL_CLAIMS,
}

LTI13_ROLE_VOCABULARIES = {
    "SYSTEM_ROLES": {
        "CORE": {
            "http://purl.imsglobal.org/vocab/lis/v2/system/person#Administrator",
            "http://purl.imsglobal.org/vocab/lis/v2/system/person#None",
        },
        "NON_CORE": {
            "http://purl.imsglobal.org/vocab/lis/v2/system/person#AccountAdmin",
            "http://purl.imsglobal.org/vocab/lis/v2/system/person#Creator",
            "http://purl.imsglobal.org/vocab/lis/v2/system/person#SysAdmin",
            "http://purl.imsglobal.org/vocab/lis/v2/system/person#SysSupport",
            "http://purl.imsglobal.org/vocab/lis/v2/system/person#User",
        },
    },
    "INSTITUTION_ROLES": {
        "CORE": {
            "http://purl.imsglobal.org/vocab/lis/v2/institution/person#Administrator",
            "http://purl.imsglobal.org/vocab/lis/v2/institution/person#Faculty",
            "http://purl.imsglobal.org/vocab/lis/v2/institution/person#Guest",
            "http://purl.imsglobal.org/vocab/lis/v2/institution/person#None",
            "http://purl.imsglobal.org/vocab/lis/v2/institution/person#Other",
            "http://purl.imsglobal.org/vocab/lis/v2/institution/person#Staff",
            "http://purl.imsglobal.org/vocab/lis/v2/institution/person#Student",
        },
        "NON_CORE": {
            "http://purl.imsglobal.org/vocab/lis/v2/institution/person#Alumni",
            "http://purl.imsglobal.org/vocab/lis/v2/institution/person#Instructor",
            "http://purl.imsglobal.org/vocab/lis/v2/institution/person#Learner",
            "http://purl.imsglobal.org/vocab/lis/v2/institution/person#Member",
            "http://purl.imsglobal.org/vocab/lis/v2/institution/person#Mentor",
            "http://purl.imsglobal.org/vocab/lis/v2/institution/person#Observer",
            "http://purl.imsglobal.org/vocab/lis/v2/institution/person#ProspectiveStudent",
        },
    },
    "CONTEXT_ROLES": {
        "CORE": {
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Administrator",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#ContentDeveloper",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Learner",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Mentor",
        },
        "NON_CORE": {
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Manager",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Member",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Officer",
        },
        "CONTEXT_SUB_ROLES": {
            # admin role
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Administrator#Administrator",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Administrator#Developer",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Administrator#ExternalDeveloper",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Administrator#ExternalSupport",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Administrator#ExternalSystemAdministrator",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Administrator#Support",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Administrator#SystemAdministrator",
            # content developer role
            "http://purl.imsglobal.org/vocab/lis/v2/membership#ContentDeveloper#ContentDeveloper",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#ContentDeveloper#ContentExpert",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#ContentDeveloper#ExternalContentExpert",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#ContentDeveloper#Librarian",
            # instructor role
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#ExternalInstructor",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#Grader",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#GuestInstructor",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#Lecturer",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#PrimaryInstructor",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#SecondaryInstructor",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#TeachingAssistant",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#TeachingAssistantGroup",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#TeachingAssistantOffering",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#TeachingAssistantSection",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#TeachingAssistantSectionAssociation",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#TeachingAssistantTemplate",
            # Learner role
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Learner#ExternalLearner",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Learner#GuestLearner",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Learner#Instructor",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Learner#Learner",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Learner#NonCreditLearner",
            # Manager role
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Manager#AreaManager",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Manager#CourseCoordinator",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Manager#ExternalObserver",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Manager#Manager",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Manager#Observer",
            # Member role
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Member#Member",
            # Mentor role
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Mentor#Mentor",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Mentor#Advisor",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Mentor#Auditor",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Mentor#ExternalAdvisor",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Mentor#ExternalAuditor",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Mentor#ExternalLearningFacilitator",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Mentor#ExternalMentor",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Mentor#ExternalReviewer",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Mentor#ExternalTutor",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Mentor#LearningFacilitator",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Mentor#Reviewer",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Mentor#Tutor",
            # Officer role
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Officer#Chair",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Officer#Communications",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Officer#Secretary",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Officer#Treasurer",
            "http://purl.imsglobal.org/vocab/lis/v2/membership#Officer#Vice-Chair",
        },
    },
}

LTI13_ROLES = {
    "STAFF_ROLES": {
        "http://purl.imsglobal.org/vocab/lis/v2/system/person#Administrator",
        "http://purl.imsglobal.org/vocab/lis/v2/system/person#None",
        "http://purl.imsglobal.org/vocab/lis/v2/institution/person#Administrator",
        "http://purl.imsglobal.org/vocab/lis/v2/institution/person#Faculty",
        "http://purl.imsglobal.org/vocab/lis/v2/membership#Administrator",
        "http://purl.imsglobal.org/vocab/lis/v2/membership#ContentDeveloper",
        "http://purl.imsglobal.org/vocab/lis/v2/membership#Manager",
    },
    "STUDENT_ROLES": {
        "http://purl.imsglobal.org/vocab/lis/v2/membership#Learner",
        "http://purl.imsglobal.org/vocab/lis/v2/membership#Learner#ExternalLearner",
        "http://purl.imsglobal.org/vocab/lis/v2/membership#Learner#Instructor",
        "http://purl.imsglobal.org/vocab/lis/v2/membership#Learner#GuestLearner",
        "http://purl.imsglobal.org/vocab/lis/v2/membership#Learner#NonCreditLearner",
        "http://purl.imsglobal.org/vocab/lis/v2/membership#Learner#Learner",
        "http://purl.imsglobal.org/vocab/lis/v2/institution/person#Student",
        "http://purl.imsglobal.org/vocab/lis/v2/institution/person#Learner",
        "http://purl.imsglobal.org/vocab/lis/v2/institution/person#ProspectiveStudent",
    },
    "INSTRUCTOR_ROLES": {
        "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor",
        "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#Grader",
        "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#TeachingAssistantOffering",
        "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#TeachingAssistantSectionAssociation",
        "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#PrimaryInstructor",
        "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#GuestInstructor",
        "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#TeachingAssistantSection",
        "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#SecondaryInstructor",
        "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#Lecturer",
        "http://purl.imsglobal.org/vocab/lis/v2/membership#Learner#Instructor",
        "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#TeachingAssistantGroup",
        "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#TeachingAssistant",
        "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#ExternalInstructor",
        "http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor#TeachingAssistantTemplate",
        "http://purl.imsglobal.org/vocab/lis/v2/institution/person#Instructor",
    },
}

# the list of roles that recognize a user as a Student
DEFAULT_ROLE_NAMES_FOR_STUDENT = ["student", "learner"]
# the list of roles that recognize a user as an Instructor
DEFAULT_ROLE_NAMES_FOR_INSTRUCTOR = [
    "instructor",
    "urn:lti:role:ims/lis/teachingassistant",
]



import os
from typing import Optional, Any

from tornado import httputil
from tornado.web import StaticFileHandler, Application

from grader_service.registry import register_handler
from grader_service.server import GraderServer


class CacheControlStaticFilesHandler(StaticFileHandler):
    """StaticFileHandler subclass that sets Cache-Control: no-cache without `?v=`

    rather than relying on default browser cache behavior.
    """

    def compute_etag(self):
        return None

    def set_extra_headers(self, path):
        if "v" not in self.request.arguments or self.settings.get(
                "no_cache_static", False
        ):
            self.add_header("Cache-Control", "no-cache")


@register_handler(r"\/logo")
class LogoHandler(StaticFileHandler):
    """A singular handler for serving the logo."""

    def __init__(self, application: "GraderServer", request: httputil.HTTPServerRequest, **kwargs: Any) -> None:
        super().__init__(application, request, **kwargs)
        self.application: GraderServer = self.application

    def initialize(self, default_filename: Optional[str] = None) -> None:
        path = self.application.logo_file
        super().initialize(path, default_filename)

    def get(self):
        return super().get('')

    @classmethod
    def get_absolute_path(cls, root, path):
        """We only serve one file, ignore relative path"""
        return os.path.abspath(root)





import typing

from nbformat import read as _read
from nbformat import reads as _reads
from nbformat import write as _write
from nbformat import writes as _writes
from nbformat.notebooknode import NotebookNode

from .common import BaseMetadataValidator, ValidationError


class MetadataValidatorV1(BaseMetadataValidator):

    schema_version = 1

    def __init__(self) -> None:
        super().__init__()

    def _upgrade_v0_to_v1(self, cell: NotebookNode) -> NotebookNode:
        meta = cell.metadata["nbgrader"]

        if "grade" not in meta:
            meta["grade"] = False
        if "solution" not in meta:
            meta["solution"] = False
        if "locked" not in meta:
            meta["locked"] = False

        if not meta["grade"] and not meta["solution"] and not meta["locked"]:
            del cell.metadata["nbgrader"]
            return cell

        if not meta["grade"]:
            if "points" in meta:
                del meta["points"]
        elif "points" in meta:
            if meta["points"] == "":
                meta["points"] = 0.0
            else:
                points = float(meta["points"])
                if points < 0:
                    meta["points"] = 0.0
                else:
                    meta["points"] = float(meta["points"])
        else:
            meta["points"] = 0.0

        meta["schema_version"] = self.schema_version

        return cell

    def upgrade_cell_metadata(self, cell: NotebookNode) -> NotebookNode:
        if "nbgrader" not in cell.metadata:
            return cell

        if "schema_version" not in cell.metadata["nbgrader"]:
            cell.metadata["nbgrader"]["schema_version"] = 0

        if cell.metadata["nbgrader"]["schema_version"] == 0:
            cell = self._upgrade_v0_to_v1(cell)

        if "nbgrader" not in cell.metadata:
            return cell

        self._remove_extra_keys(cell)
        return cell

    def validate_cell(self, cell: NotebookNode) -> None:
        super(MetadataValidatorV1, self).validate_cell(cell)

        if "nbgrader" not in cell.metadata:
            return

        meta = cell.metadata["nbgrader"]
        grade = meta["grade"]
        solution = meta["solution"]
        locked = meta["locked"]

        # check for a valid grade id
        if grade or solution or locked:
            if "grade_id" not in meta:
                raise ValidationError(
                    "nbgrader cell does not have a grade_id: {}".format(cell.source)
                )
            if meta["grade_id"] == "":
                raise ValidationError("grade_id is empty")

        # check for valid points
        if grade:
            if "points" not in meta:
                raise ValidationError(
                    "nbgrader cell '{}' does not have points".format(meta["grade_id"])
                )

        # check that markdown cells are grade AND solution (not either/or)
        if cell.cell_type == "markdown" and grade and not solution:
            raise ValidationError(
                "Markdown grade cell '{}' is not marked as a solution cell".format(
                    meta["grade_id"]
                )
            )
        if cell.cell_type == "markdown" and not grade and solution:
            raise ValidationError(
                "Markdown solution cell is not marked as a grade cell: {}".format(
                    cell.source
                )
            )

    def validate_nb(self, nb: NotebookNode) -> None:
        super(MetadataValidatorV1, self).validate_nb(nb)

        ids = set([])
        for cell in nb.cells:

            if "nbgrader" not in cell.metadata:
                continue

            grade = cell.metadata["nbgrader"]["grade"]
            solution = cell.metadata["nbgrader"]["solution"]
            locked = cell.metadata["nbgrader"]["locked"]

            if not grade and not solution and not locked:
                continue

            grade_id = cell.metadata["nbgrader"]["grade_id"]
            if grade_id in ids:
                raise ValidationError("Duplicate grade id: {}".format(grade_id))
            ids.add(grade_id)


def read_v1(
    source: typing.TextIO, as_version: int, **kwargs: typing.Any
) -> NotebookNode:
    nb = _read(source, as_version, **kwargs)
    MetadataValidatorV1().validate_nb(nb)
    return nb


def write_v1(nb: NotebookNode, fp: typing.TextIO, **kwargs: typing.Any) -> None:
    MetadataValidatorV1().validate_nb(nb)
    _write(nb, fp, **kwargs)


def reads_v1(source: str, as_version: int, **kwargs: typing.Any) -> NotebookNode:
    nb = _reads(source, as_version, **kwargs)
    MetadataValidatorV1().validate_nb(nb)
    return nb


def writes_v1(nb: NotebookNode, **kwargs: typing.Any) -> None:
    MetadataValidatorV1().validate_nb(nb)
    _writes(nb, **kwargs)



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import enum

from grader_service.api.models.base_model import Model
from sqlalchemy.orm import declarative_base

Base = declarative_base()


class Serializable(object):
    @property
    def model(self) -> Model:
        return Model()

    def serialize(self) -> dict:
        return self.model.to_dict()


class DeleteState(enum.IntEnum):
    active = 0
    deleted = 1



"""HTTP Handlers for the hub server"""
from typing import Optional, Awaitable

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from jinja2 import Template
from tornado import web
from tornado.escape import url_escape
from tornado.httputil import url_concat

from grader_service.handlers.base_handler import BaseHandler


class LogoutHandler(BaseHandler):
    """Log a user out by clearing their login cookie."""


    def _backend_logout_cleanup(self, name):
        """Default backend logout actions
        """
        self.log.info("User logged out: %s", name)
        self.clear_login_cookies()

    async def default_handle_logout(self):
        """The default logout action

        Optionally cleans up servers, clears cookies, increments logout counter
        Cleaning up servers can be prevented by setting shutdown_on_logout to
        False.
        """
        user = self.current_user
        if user:
            self._backend_logout_cleanup(user.name)

    async def handle_logout(self):
        """Custom user action during logout

        By default a no-op, this function should be overridden in subclasses
        to have JupyterHub take a custom action on logout.
        """
        return

    async def render_logout_page(self):
        """Render the logout page, if any

        Override this function to set a custom logout page.
        """
        if self.authenticator.auto_login:
            html = await self.render_template('auth/logout.html.j2')
            self.finish(html)
        else:
            self.redirect(self.settings['login_url'], permanent=False)

    async def get(self):
        """Log the user out, call the custom action, forward the user
        to the logout page
        """
        await self.default_handle_logout()
        await self.handle_logout()
        # clear grader user before rendering logout page
        # ensures the login button is shown instead of logout
        self._grader_user = None
        await self.render_logout_page()


class LoginHandler(BaseHandler):
    """Render the login page."""

    def _render(self, login_error=None, username=None):
        context = {
            "next": url_escape(self.get_argument('next', default='')),
            "username": username,
            "login_error": login_error,
            "login_url": self.settings['login_url'],
            "authenticator_login_url": url_concat(
                self.authenticator.login_url(self.application.base_url),
                {
                    'next': self.get_argument('next', ''),
                },
            ),
            "authenticator": self.authenticator,
            "xsrf": self.xsrf_token.decode('ascii'),
        }
        custom_html = Template(
            self.authenticator.get_custom_html(self.application.base_url)
        ).render(**context)
        return self.render_template(
            'auth/login.html.j2',
            **context,
            custom_html=custom_html,
        )

    async def prepare(self) -> Optional[Awaitable[None]]:
        await super().prepare()
        return

    async def get(self):
        user = self.current_user
        if user:
            # set new login cookie
            # because single-user cookie may have been cleared or incorrect
            self.set_login_cookie(user)
            self.redirect(self.get_next_url(user), permanent=False)
        else:
            if self.authenticator.auto_login:
                auto_login_url = self.authenticator.login_url(self.application.base_url)
                if auto_login_url == self.settings['login_url']:
                    # auto_login without a custom login handler
                    # means that auth info is already in the request
                    # (e.g. REMOTE_USER header)
                    user = await self.login_user()
                    if user is None:
                        # auto_login failed, just 403
                        raise web.HTTPError(403)
                    else:
                        self.redirect(self.get_next_url(user))
                else:
                    if self.get_argument('next', default=False):
                        auto_login_url = url_concat(
                            auto_login_url, {'next': self.get_next_url()}
                        )
                    self.redirect(auto_login_url)
                return
            username = self.get_argument('username', default='')
            self.finish(await self._render(username=username))

    async def post(self):
        # parse the arguments dict
        data = {}
        for arg in self.request.body_arguments:
            if arg == "_xsrf":
                # don't include xsrf token in auth input
                continue
            # strip username, but not other fields like passwords,
            # which should be allowed to start or end with space
            data[arg] = self.get_argument(arg, strip=arg == "username")

        user = await self.login_user(data)

        if user:
            # register current user for subsequent requests to user (e.g. logging the request)
            self._grader_user = user
            self.redirect(self.get_next_url(user))
        else:
            html = await self._render(
                login_error='Invalid username or password', username=data['username']
            )
            await self.finish(html)


